###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test class for ModuleBase."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, missing-docstring, invalid-name

__RCSID__ = "$Id$"

import os
from itertools import product
from mock import MagicMock

import pytest

# mocks
from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock
from LHCbDIRAC.Workflow.Modules.mock_Commons import version, prod_id, prod_job_id, wms_job_id, \
    workflowStatus, stepStatus, step_id, step_number,\
    step_commons, wf_commons

# sut
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


def test__enableModule(mocker):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
  mb.execute(version, prod_id, prod_job_id, wms_job_id,
             workflowStatus, stepStatus,
             wf_commons, step_commons[0],
             step_number, step_id)
  assert mb._enableModule() is True


def test__checkLocalExistance(mocker):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
  with pytest.raises(OSError):
    mb._checkLocalExistance(['aaa', 'bbb'])

  with open('testFile.txt', 'w'):
    res = mb._checkLocalExistance(['testFile.txt'])
    assert res == ['testFile.txt']

    with open('testSecondFile.txt', 'w'):
      res = mb._checkLocalExistance(['testFile.txt', 'testSecondFile.txt'])
      assert sorted(res) == sorted(['testFile.txt', 'testSecondFile.txt'])

      res = mb._checkLocalExistance(['TESTFILE.TXT', 'testSecondFile.txt'])
      assert sorted(res) == sorted(['testFile.txt', 'testSecondFile.txt'])
  os.remove('testFile.txt')
  os.remove('testSecondFile.txt')


candidateFiles = {'00012345_00012345_4.dst': {'lfn': '/lhcb/MC/2010/DST/123/123_45_4.dst',
                                              'type': 'dst'},
                  '00012345_00012345_2.digi': {'type': 'digi'},
                  '00012345_00012345_3.digi': {'type': 'digi'},
                  '00012345_00012345_5.AllStreams.dst': {'lfn': '/lhcb/MC/2010/DST/123/123_45_5.AllStreams.dst',
                                                         'type': 'allstreams.dst'},
                  '00012345_00012345_1.sim': {'type': 'sim'},
                  '00038941_00000004_6.B2_D2.Strip.dst': {'lfn': '/lhcb/MC/2012/B2_D2.STRIP.DST/B2_D2.Strip.dst',
                                                          'type': 'B2_D2.strip.dst'},
                  'Gauss_HIST_1.root': {'type': 'GAUSSHIST'}}

fileMasks = (['dst'], 'dst', ['sim'], ['digi'], ['digi', 'sim'], 'allstreams.dst',
             'B2_D2.strip.dst', [],
             ['B2_D2.strip.dst', 'digi'],
             ['gausshist', 'digi'])
stepMasks = ('', '5', '', ['2'], ['1', '3'], '',
             '', ['6'],
             [],
             ['1', '3'], )

results = ({'00012345_00012345_4.dst': {'lfn': '/lhcb/MC/2010/DST/123/123_45_4.dst',
                                        'type': 'dst'}},
           {},
           {'00012345_00012345_1.sim': {'type': 'sim'}},
           {'00012345_00012345_2.digi': {'type': 'digi'}, },
           {'00012345_00012345_3.digi': {'type': 'digi'},
            '00012345_00012345_1.sim': {'type': 'sim'}},
           {'00012345_00012345_5.AllStreams.dst':
            {'lfn': '/lhcb/MC/2010/DST/123/123_45_5.AllStreams.dst',
             'type': 'allstreams.dst'}},
           {'00038941_00000004_6.B2_D2.Strip.dst':
            {'lfn': '/lhcb/MC/2012/B2_D2.STRIP.DST/B2_D2.Strip.dst',
             'type': 'B2_D2.strip.dst'}},
           {'00038941_00000004_6.B2_D2.Strip.dst':
            {'lfn': '/lhcb/MC/2012/B2_D2.STRIP.DST/B2_D2.Strip.dst',
             'type': 'B2_D2.strip.dst'}},
           {'00012345_00012345_2.digi': {'type': 'digi'},
            '00012345_00012345_3.digi': {'type': 'digi'},
            '00038941_00000004_6.B2_D2.Strip.dst':
            {'lfn': '/lhcb/MC/2012/B2_D2.STRIP.DST/B2_D2.Strip.dst',
             'type': 'B2_D2.strip.dst'}},
           {'00012345_00012345_3.digi': {'type': 'digi'},
            'Gauss_HIST_1.root': {'type': 'GAUSSHIST'}},)

allCombinations = list(zip(fileMasks, stepMasks, results))


@pytest.mark.parametrize("fileMask, stepMask, result", allCombinations)
def test__applyMask(mocker, fileMask, stepMask, result):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
  res = mb._applyMask(candidateFiles, fileMask, stepMask)
  assert res == result


def test__checkSanity(mocker):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
  with pytest.raises(ValueError):
    mb._checkSanity(candidateFiles)


outputList = [{'outputDataType': 'txt', 'outputDataName': 'foo_1.txt'},
              {'outputDataType': 'py', 'outputDataName': 'bar_2.py'},
              {'outputDataType': 'blah', 'outputDataName': 'myfoo.blah'}]
outputLFNs = ['/lhcb/MC/2010/DST/00012345/0001/foo_1.txt',
              '/lhcb/MC/2010/DST/00012345/0001/bar_2.py',
              '/lhcb/user/f/fstagni/2015_06/prepend_myfoo.blah']

fileMasks = ['txt', ['txt', 'py'], ['blah'], ['aa'], '', '', '']
stepMasks = ['', None, None, None, '2', ['2', '3'], ['3']]
results = [{'foo_1.txt': {'lfn': '/lhcb/MC/2010/DST/00012345/0001/foo_1.txt',
                          'type': outputList[0]['outputDataType']}},
           {'foo_1.txt': {'lfn': '/lhcb/MC/2010/DST/00012345/0001/foo_1.txt',
                          'type': outputList[0]['outputDataType']},
            'bar_2.py': {'lfn': '/lhcb/MC/2010/DST/00012345/0001/bar_2.py',
                         'type': outputList[1]['outputDataType']}},
           {'myfoo.blah': {'lfn': '/lhcb/user/f/fstagni/2015_06/prepend_myfoo.blah',
                           'type': outputList[2]['outputDataType']}},
           {},
           {'bar_2.py': {'lfn': '/lhcb/MC/2010/DST/00012345/0001/bar_2.py',
                         'type': outputList[1]['outputDataType']}},
           {'bar_2.py': {'lfn': '/lhcb/MC/2010/DST/00012345/0001/bar_2.py',
                         'type': outputList[1]['outputDataType']}},
           {}]

allCombinations = list(zip(fileMasks, stepMasks, results))


@pytest.mark.parametrize("fileMask, stepMask, result", allCombinations)
def test_getCandidateFiles(mocker, fileMask, stepMask, result):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)

  with open('foo_1.txt', 'w'):
    with open('bar_2.py', 'w'):
      with open('myfoo.blah', 'w'):
        res = mb.getCandidateFiles(outputList, outputLFNs, fileMask, stepMask)
        assert res == result

        with pytest.raises(ValueError):
          mb.getCandidateFiles([{'outputDataType': 'txt', 'outputDataName': 'foo_1.txt'},
                                {'outputDataType': 'py', 'outputDataName': 'bar_2.py'},
                                {'outputDataType': 'blah', 'outputDataName': 'myfoo.blah'}],
                               ['/lhcb/MC/2010/DST/00012345/0001/foo_1.txt',
                                '/lhcb/MC/2010/DST/00012345/0001/bar_2.py',
                                '/lhcb/user/f/fstagni/2015_06/prepend_myfoo_1.blah'],
                               'blah',
                               '')
        os.remove('foo_1.txt')
        os.remove('bar_2.py')
        os.remove('myfoo.blah')


inputDatas = ['previousStep', 'previousStep', 'LFN:123.raw']
workflow_commons = [{'JobType': 'User',
                     'outputList': [{'stepName': 'Brunel_1',
                                     'outputDataType': 'brunelhist',
                                     'outputBKType': 'BRUNELHIST',
                                     'outputDataName': 'Brunel_00012345_00006789_1_Hist.root'},
                                    {'stepName': 'Brunel_1',
                                     'outputDataType': 'sdst',
                                     'outputBKType': 'SDST',
                                     'outputDataName': '00012345_00006789_1.sdst'}
                                    ]
                     },
                    {'JobType': 'User',
                     'outputList': [{'stepName': 'Brunel_1',
                                     'outputDataType': 'brunelhist',
                                     'outputBKType': 'BRUNELHIST',
                                     'outputDataName': 'Brunel_00012345_00006789_1_Hist.root'},
                                    {'stepName': 'Brunel_1',
                                     'outputDataType': 'sdst',
                                     'outputBKType': 'SDST',
                                     'outputDataName': 'some.sdst'},
                                    {'stepName': 'Brunel_1',
                                     'outputDataType': 'sdst',
                                     'outputBKType': 'SDST',
                                     'outputDataName': '00012345_00006789_1.sdst'}
                                    ]
                     },
                    {'JobType': 'User',
                     'outputList': [{'stepName': 'Brunel_1',
                                     'outputDataType': 'brunelhist',
                                     'outputBKType': 'BRUNELHIST',
                                     'outputDataName': 'Brunel_00012345_00006789_1_Hist.root'},
                                    {'stepName': 'Brunel_1',
                                     'outputDataType': 'sdst',
                                     'outputBKType': 'SDST',
                                     'outputDataName': 'some.sdst'},
                                    {'stepName': 'Brunel_1',
                                     'outputDataType': 'sdst',
                                     'outputBKType': 'SDST',
                                     'outputDataName': '00012345_00006789_1.sdst'}
                                    ]
                     }
                    ]
results = [['00012345_00006789_1.sdst'],
           ['some.sdst', '00012345_00006789_1.sdst'],
           ['123.raw']]

allCombinations = list(zip(inputDatas, workflow_commons, results))


@pytest.mark.parametrize("inputData, wf_c, result", allCombinations)
def test__determineStepInputData(mocker, inputData, wf_c, result):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
  mb.stepName = 'DaVinci_2'

  mb.gaudiSteps = ['Brunel_1', 'DaVinci_2']
  mb.workflow_commons = wf_c
  mb.inputDataType = 'SDST'
  res = mb._determineStepInputData(inputData)
  assert res == result


@pytest.mark.parametrize("s_cs", list(step_commons))
def test__determineOutputs(mocker, s_cs):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
  mb.jobType = 'merge'
  mb.step_id = '00000123_00000456_1'
  mb.stepInputData = ['foo', 'bar']
  mb.step_commons = dict(s_cs)
  mb.step_commons['listoutput'] = [{'outputDataType': 'bhadron.dst;sdst',
                                    'outputDataName': '00000123_00000456_1.bhadron.dst;sdst'}]
  outF, outft, histos = mb._determineOutputs()
  assert outF == [{'outputDataType': 'sdst',
                   'outputDataName': '00000123_00000456_1.sdst',
                   'outputBKType': 'SDST'}]
  assert outft == ['sdst']
  assert histos is False

  mb.step_commons['listoutput'] = [{'outputDataType': 'root',
                                    'outputDataName': '00000123_00000456_1.root',
                                    'outputBKType': 'ROOT'}]
  outF, outft, histos = mb._determineOutputs()
  assert outF == [{'outputDataType': 'root',
                   'outputDataName': '00000123_00000456_1.root',
                   'outputBKType': 'ROOT'}]
  assert outft == ['root']
  assert histos is False

  mb.jobType = 'reco'
  mb.step_commons = s_cs
  mb.step_commons['listoutput'] = [{'outputDataType': 'sdst',
                                    'outputDataName': '00000123_00000456_1.sdst',
                                    'outputBKType': 'SDST'}]
  outF, outft, histos = mb._determineOutputs()
  assert outF == [{'outputDataType': 'sdst',
                   'outputDataName': '00000123_00000456_1.sdst',
                   'outputBKType': 'SDST'}]
  assert outft == ['sdst']
  assert histos is False


def test__findOutputs(mocker):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)

  with pytest.raises(IOError):
    with open('aaa.Bhadron.dst', 'w'):
      with open('ccc.charm.mdst', 'w'):
        with open('prova.txt', 'w'):
          stepOutput = [
              {
                  'outputDataType': 'BHADRON.DST', 'outputDataName': 'aaa.bhadron.dst'}, {
                  'outputDataType': 'CALIBRATION.DST', 'outputDataName': 'bbb.calibration.dst'}, {
                  'outputDataType': 'CHARM.MDST', 'outputDataName': 'ccc.charm.mdst'}, {
                  'outputDataType': 'CHARMCONTROL.DST', 'outputDataName': '00012345_00012345_2.CHARMCONTROL.DST'}, {
                  'outputDataType': 'CHARMFULL.DST', 'outputDataName': '00012345_00012345_2.CHARMFULL.DST'}, {
                  'outputDataType': 'LEPTONIC.MDST', 'outputDataName': '00012345_00012345_2.LEPTONIC.MDST'}, {
                  'outputDataType': 'LEPTONICFULL.DST', 'outputDataName': '00012345_00012345_2.LEPTONICFULL.DST'}, {
                  'outputDataType': 'MINIBIAS.DST', 'outputDataName': '00012345_00012345_2.MINIBIAS.DST'}, {
                  'outputDataType': 'RADIATIVE.DST', 'outputDataName': '00012345_00012345_2.RADIATIVE.DST'}, {
                  'outputDataType': 'SEMILEPTONIC.DST', 'outputDataName': '00012345_00012345_2.SEMILEPTONIC.DST'}, {
                  'outputDataType': 'HIST', 'outputDataName': 'DaVinci_00012345_00012345_2_Hist.root'}]
          mb._findOutputs(stepOutput)
          os.remove('aaa.Bhadron.dst')
          os.remove('ccc.charm.mdst')
          os.remove('prova.txt')

  stepOutput = [{'outputDataType': 'BHADRON.DST', 'outputDataName': 'aaa.bhadron.dst'}]
  outExp = [{'outputDataType': 'bhadron.dst', 'outputBKType': 'BHADRON.DST', 'outputDataName': 'aaa.Bhadron.dst',
             'stepName': 'someApp_1'}]
  bkExp = ['BHADRON.DST']
  mb.stepName = 'someApp_1'
  out, bk = mb._findOutputs(stepOutput)
  assert out == outExp
  assert bk == bkExp


def test_getFileMetadata(mocker):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)

  candidateFiles = {'foo_1.txt': {'lfn': '/lhcb/MC/2010/DST/00012345/0001/foo_1.txt',
                                  'type': 'txt',
                                  'workflowSE': 'SE1'},
                    'bar_2.py': {'lfn': '/lhcb/MC/2010/DST/00012345/0001/bar_2.py',
                                 'type': 'py',
                                 'workflowSE': 'SE2'},
                    }

  expectedResult = {'bar_2.py': {'filedict': {'Status': 'Waiting',
                                              'LFN': '/lhcb/MC/2010/DST/00012345/0001/bar_2.py',
                                              'GUID': 'D41D8CD9-8F00-B204-E980-0998ECF8427E',
                                              'Checksum': '000001',
                                              'ChecksumType': 'ADLER32',
                                              'Size': 0},
                                 'lfn': '/lhcb/MC/2010/DST/00012345/0001/bar_2.py',
                                 'workflowSE': 'SE2',
                                 'localpath': os.getcwd() + '/bar_2.py',
                                 'guid': 'D41D8CD9-8F00-B204-E980-0998ECF8427E',
                                 'type': 'py'},
                    'foo_1.txt': {'filedict': {'Status': 'Waiting',
                                               'LFN': '/lhcb/MC/2010/DST/00012345/0001/foo_1.txt',
                                               'GUID': 'D41D8CD9-8F00-B204-E980-0998ECF8427E',
                                               'Checksum': '000001',
                                               'ChecksumType': 'ADLER32',
                                               'Size': 0},
                                  'lfn': '/lhcb/MC/2010/DST/00012345/0001/foo_1.txt',
                                  'workflowSE': 'SE1',
                                  'localpath': os.getcwd() + '/foo_1.txt',
                                  'guid': 'D41D8CD9-8F00-B204-E980-0998ECF8427E',
                                  'type': 'txt'}
                    }

  with open('foo_1.txt', 'w'):
    with open('bar_2.py', 'w'):
      result = mb.getFileMetadata(candidateFiles)
      assert result == expectedResult
      os.remove('foo_1.txt')
      os.remove('bar_2.py')


allCombinations = product(list(workflow_commons), step_commons)


@pytest.mark.parametrize("wf_c, s_cs", allCombinations)
def test_createProdConfFile(mocker, wf_c, s_cs):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  mb = ModuleBase(bkClientIn=bkc_mock, dm=dm_mock)
  mb.workflow_commons = wf_c
  mb.step_commons = s_cs
  mb._resolveInputVariables()
  mb._resolveInputStep()
  res = mb.createProdConfFile(['DST', 'GAUSSHIST'], True, 123, 1)
  print(res)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Fix incorrect file GUIDs."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript

keep = None
localFile = None


def leave(msg, error=None, exitCode=0):
  from DIRAC import gLogger

  if error is None:
    error = {}

  if not keep and os.path.exists(localFile):
    os.remove(localFile)
    gLogger.info("Local file %s removed" % localFile)

  if error:
    errMsg = error.get(
        'Message',
        list(error.get('Value', {}).get('Failed', {'': 'Unknown reason'}).values())[0],
    )
    gLogger.error(msg, ': %s' % errMsg)
  else:
    gLogger.always(msg)
  DIRAC.exit(exitCode)


@DIRACScript()
def main():
  global keep
  global localFile

  from DIRAC.Core.Base import Script
  Script.registerSwitch("f:", "OldLFN=", "LFN of existing file to be fixed.")
  Script.registerSwitch("n:", "NewLFN=",
                        "Optional: specify a new LFN for the file (retaining the existing file with incorrect GUID).")
  Script.registerSwitch("D:", "Directory=", "Optional: directory to download file (defaults to TMPDIR then PWD).")
  Script.registerSwitch("k", "Keep", "Optional: specify this switch to retain the local copy of the downloaded file")
  Script.registerSwitch("m", "SafeMode",
                        "Optional: specify this switch to run the script in safe mode (will check the GUIDs only)")
  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile] [OldLFN]' % Script.scriptName]))
  Script.parseCommandLine(ignoreErrors=True)

  args = Script.getPositionalArgs()

  from DIRAC.Interfaces.API.Dirac import Dirac
  from LHCbDIRAC.Core.Utilities.File import makeGuid

  from DIRAC import gLogger

  oldLFN = ''
  newLFN = ''
  localFile = ''
  keep = False
  safe = False
  directory = os.getcwd()
  if 'TMPDIR' in os.environ:
    directory = os.environ['TMPDIR']

  dirac = Dirac()

  if args:
    oldLFN = args.pop(0)

  for switch in Script.getUnprocessedSwitches():
    if switch[0].lower() in ('f', 'oldlfn'):
      oldLFN = switch[1]
    elif switch[0].lower() in ('n', 'newlfn'):
      newLFN = switch[1]
    elif switch[0].lower() in ('d', 'directory'):
      directory = switch[1]
    elif switch[0].lower() in ('k', 'keep'):
      keep = True
    elif switch[0].lower() in ('m', 'safemode'):
      safe = True

  oldLFN = oldLFN.replace('LFN:', '')
  newLFN = newLFN.replace('LFN:', '')

  if not oldLFN:
    leave('The original LFN of the file to be fixed must be specified', exitCode=2)

  if not os.path.exists(directory):
    leave("Optional directory %s doesn't exist" % directory, exitCode=2)

  if not newLFN:
    gLogger.verbose('No new LFN specified, will replace the existing LFN %s' % (oldLFN))
    newLFN = oldLFN

  gLogger.verbose('Directory for downloading file is set to %s' % directory)

  replicaInfo = dirac.getReplicas(oldLFN)
  if not replicaInfo['OK'] or replicaInfo['Value']['Failed']:
    leave('Could not get replica information for %s' % oldLFN, replicaInfo, exitCode=2)

  replicas = replicaInfo['Value']['Successful'][oldLFN]
  storageElements = list(replicas)
  if not storageElements:
    leave('Could not determine SEs for replicas of %s' % oldLFN, exitCode=2)

  gLogger.info('Existing LFN has replicas at: %s' % ', '.join(storageElements))

  oldGUID = dirac.getLfnMetadata(oldLFN)
  if not oldGUID['OK'] or oldGUID['Value']['Failed']:
    leave('Could not obtain GUID from LFC for %s - %s' % oldLFN, oldGUID, exitCode=2)
  oldGUID = oldGUID['Value']['Successful'][oldLFN]['GUID']
  gLogger.verbose('Existing GUID is %s' % oldGUID)

  # retrieve original file
  localFile = os.path.join(directory, os.path.basename(oldLFN))
  if not os.path.exists(localFile):
    download = dirac.getFile(oldLFN, directory)
    if not download['OK'] or download['Value']['Failed']:
      leave('Could not download file with message - %s' % download['Message'], download, exitCode=2)
  else:
    gLogger.always('Found file %s in local directory, will not re-download' % os.path.basename(oldLFN))

  newGUID = makeGuid(localFile)[localFile]

  if newGUID == oldGUID:
    leave('Old and new GUIDs have the same value (%s), exiting without changes' % oldGUID)

  if safe:
    leave('Safe mode - found file GUID = %s and existing GUID = %s, exiting without changes' % (newGUID, oldGUID))

  gLogger.verbose('Will set old GUID to %s from %s' % (newGUID, oldGUID))
  if newLFN == oldLFN:
    gLogger.always('Removing old LFN from storages before adding new LFN')
    result = dirac.removeFile(oldLFN)
    if not result['OK']:
      leave('Could not remove existing LFN from Grid storage', result, exitCode=2)

  gLogger.always('Uploading %s as LFN:%s with replica at %s and GUID = %s' % (localFile, newLFN,
                                                                              storageElements[0], newGUID))
  result = dirac.addFile(newLFN, localFile, storageElements[0], fileGuid=newGUID, printOutput=False)
  if not result['OK']:
    leave('Failed to copy and register new LFN:%s' % newLFN, result, exitCode=2)

  leave('')


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test the logic to detect successful and failed file accesses via xroot.

We use test pool and summary xml files that correspond to 3 LFNs:
* one that worked at the first attempt
* one that worked at the second attempt
* one that did not work at all
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from collections import defaultdict
from mock import MagicMock, patch

from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary
from DIRAC.Resources.Catalog.PoolXMLCatalog import PoolXMLCatalog

from LHCbDIRAC.Workflow.Modules.AnalyseFileAccess import AnalyseFileAccess

testdir = os.path.dirname(__file__)
poolFile = os.path.join(testdir, 'pool_xml_catalog.xml')
summaryFile = os.path.join(testdir, 'summary.xml')


@patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
def test_analyseFileAccess(mockRequestValidator):
  """Analyze the file accesses from a pool xml catalog and the xml summary."""

  xmlCatalog = PoolXMLCatalog(xmlfile=poolFile)
  xmlSummary = XMLSummary(summaryFile)

  fileAccessAnalyzer = AnalyseFileAccess()
  accessAttempts = fileAccessAnalyzer._checkFileAccess(xmlCatalog, xmlSummary)

  # Count how many successes
  accessPerSE = defaultdict(lambda: defaultdict(int))

  for se, success in accessAttempts:
    accessPerSE[se][success] += 1

  # 'NEVERUSED-DST' should not have any counter
  assert 'NEVERUSED-DST' not in accessPerSE

  # GOOD-DST should have 2 good reads and no bad
  assert False not in accessPerSE['GOOD-DST']
  assert accessPerSE['GOOD-DST'][True] == 2

  # BAD-DST should have 2 bad reads and no good
  assert True not in accessPerSE['BAD-DST']
  assert accessPerSE['BAD-DST'][False] == 2

  # OTHERBAD-DST should have no good read and 1 bad read
  assert True not in accessPerSE['OTHERBAD-DST']
  assert accessPerSE['OTHERBAD-DST'][False] == 1



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from mock import Mock
import unittest

from LHCbDIRAC.TransformationSystem.Utilities.StateMachine import TransformationFilesStateMachine


class UtilitiesTestCase(unittest.TestCase):

  def setUp(self):
    self.transClient = Mock()

  def tearDown(self):
    pass


class tsfmSuccess(UtilitiesTestCase):
  def test_setState(self):
    tsfm = TransformationFilesStateMachine('Unused')
    res = tsfm.setState('Assigned')
    self.assertTrue(res['OK'])
    self.assertEqual(res['Value'], 'Assigned')

    tsfm = TransformationFilesStateMachine('Unused')
    res = tsfm.setState('Unused')
    self.assertTrue(res['OK'])
    self.assertEqual(res['Value'], 'Unused')

    tsfm = TransformationFilesStateMachine('Unused')
    res = tsfm.setState('MissingInFC')
    self.assertTrue(res['OK'])
    self.assertEqual(res['Value'], 'MissingInFC')

    tsfm = TransformationFilesStateMachine('Unused')
    res = tsfm.setState('Processed')
    self.assertTrue(res['OK'])
    self.assertEqual(res['Value'], 'Processed')

    tsfm = TransformationFilesStateMachine('Unused')
    res = tsfm.setState('Processed')
    self.assertTrue(res['OK'])
    self.assertEqual(res['Value'], 'Processed')
    res = tsfm.setState('Processed')
    self.assertTrue(res['OK'])
    self.assertEqual(res['Value'], 'Processed')


class tsfmFailure(UtilitiesTestCase):
  def test_setState(self):
    tsfm = TransformationFilesStateMachine('Unused')
    res = tsfm.setState('Assigne')
    self.assertFalse(res['OK'])

    tsfm = TransformationFilesStateMachine('Processed')
    res = tsfm.setState('Assigne')
    self.assertFalse(res['OK'])

    tsfm = TransformationFilesStateMachine('Processed')
    res = tsfm.setState('Processed')
    self.assertTrue(res['OK'])

    tsfm = TransformationFilesStateMachine('Processed')
    res = tsfm.setState('Unused')
    self.assertTrue(res['OK'])
    self.assertEqual(res['Value'], 'Processed')

    tsfm = TransformationFilesStateMachine('MaxReset')
    res = tsfm.setState('Processed')
    self.assertTrue(res['OK'])
    self.assertEqual(res['Value'], 'MaxReset')


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(UtilitiesTestCase)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(tsfmSuccess))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for Workflow Module Uploads."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, missing-docstring

__RCSID__ = "$Id$"

import os
import copy
import json

from mock import MagicMock
import pytest

from DIRAC import gLogger
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File

# mocks
from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from DIRAC.Resources.Catalog.test.mock_FC import fc_mock

from LHCbDIRAC.Workflow.Modules.mock_Commons import prod_id, prod_job_id, wms_job_id, \
    workflowStatus, stepStatus, step_id, step_number,\
    step_commons, wf_commons,\
    rc_mock
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock

# sut
from LHCbDIRAC.Workflow.Modules.UploadOutputData import UploadOutputData


ft_mock = MagicMock()
ft_mock.transferAndRegisterFile.return_value = {'OK': True, 'Value': {'uploadedSE': ''}}
ft_mock.transferAndRegisterFileFailover.return_value = {'OK': True, 'Value': {}}
ft_mock.request = rc_mock
ft_mock.FileCatalog = fc_mock


@pytest.fixture
def rmFiles():
  yield
  for fileProd in ['appLog',
                   'foo.txt',
                   'bar.txt',
                   'aaa.Bhadron.dst',
                   'bbb.Calibration.dst',
                   'bar.py',
                   'aLongLog.log',
                   'bookkeeping_123_00000456_321.xml',
                   'aLongLog.log.gz',
                   'ccc.charm.mdst',
                   'ccc.charm.mdst',
                   'prova.txt',
                   'aLog.log',
                   'BAR.txt',
                   'FooBAR.ext.txt',
                   'foo_1.txt',
                   'bar_2.py',
                   'ErrorLogging_Step1_coredump.log',
                   '123_00000456_request.xml',
                   'lfn1',
                   'lfn2',
                   'XMLSummaryFile',
                   'aaa.bhadron.dst',
                   'bbb.calibration.dst',
                   'ProductionOutputData',
                   'data.py',
                   '123_00000456_request.json',
                   '00000123_00000456.tar',
                   'someOtherDir',
                   'DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK',
                   'myfoo.blah',
                   'prodConf_someApp__.py',
                   'prodConf_someApp___.py']:
    try:
      os.remove(fileProd)
    except OSError:
      continue


def test_execute(mocker, rmFiles):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
  mocker.patch("LHCbDIRAC.Workflow.Modules.UploadOutputData.FileCatalog", side_effect=fc_mock)
  mocker.patch("LHCbDIRAC.Core.Utilities.ResolveSE.gConfig", side_effect=MagicMock())

  uod = UploadOutputData(bkClient=bkc_mock, dm=dm_mock)
  uod.log = gLogger
  uod.log.setLevel('DEBUG')
  uod.siteName = 'DIRAC.Test.ch'
  uod.failoverTransfer = ft_mock

  # no errors, no input data
  for wf_cs in copy.deepcopy(wf_commons[0:1]):
    if 'outputList' in wf_cs:
      continue
    for s_cs in step_commons[0:1]:
      fileDescendants = {}
      assert uod.execute(prod_id, prod_job_id, wms_job_id,
                         workflowStatus, stepStatus,
                         wf_cs, s_cs,
                         step_number, step_id,
                         SEs=['SomeSE'],
                         fileDescendants=fileDescendants)['OK'] is True

  # no errors, input data
  for wf_cs in copy.deepcopy(wf_commons):
    for s_cs in step_commons:
      for transferAndRegisterFile in ({'OK': True, 'Value': {'uploadedSE': ''}}, {'OK': False, 'Message': 'error'}):
        # for transferAndRegisterFileFailover in ( {'OK': True, 'Value': {}},
        # {'OK': False, 'Message': 'error'} ):
        ft_mock.transferAndRegisterFile.return_value = transferAndRegisterFile
        open('foo.txt', 'w').close()
        open('bar.txt', 'w').close()
        if 'InputData' not in wf_cs:
          continue
        if wf_cs['InputData'] == '':
          continue
        wf_cs['outputList'] = [{'outputDataType': 'txt', 'outputDataName': 'foo.txt'},
                               {'outputDataType': 'txt', 'outputDataName': 'bar.txt'},
                               ]
        wf_cs['ProductionOutputData'] = ['/lhcb/MC/2010/DST/00012345/0001/foo.txt',
                                         '/lhcb/MC/2010/DST/00012345/0001/bar.txt']
#          bkc_mock.getFileDescendants.return_value = {'OK': False,
#                                                           'rpcStub': ( ( 'Bookkeeping/BookkeepingManager',
#                                                                        {'skipCACheck': False,
#                                                                         'timeout': 3600} ),
#                                                                       'getFileDescendants', ( ['foo'], 9, 0, True ) ),
#                                                           'Value': {'Successful': {'foo.txt': ['baaar']},
#                                                                     'Failed': [],
#                                                                     'NotProcessed': []}}
        fileDescendants = {'foo.txt': ['baaar']}
        assert uod.execute(prod_id, prod_job_id, wms_job_id,
                           workflowStatus, stepStatus,
                           wf_cs, s_cs,
                           step_number, step_id,
                           SEs=['SomeSE'],
                           fileDescendants=fileDescendants)['OK'] is False

#          bkc_mock.getFileDescendants.return_value = {'OK': True,
#                                                           'rpcStub': ( ( 'Bookkeeping/BookkeepingManager',
#                                                                        {'skipCACheck': False,
#                                                                         'timeout': 3600} ),
#                                                                       'getFileDescendants', ( ['foo'], 9, 0, True ) ),
#                                                           'Value': {'Successful': {},
#                                                                     'Failed': [],
#                                                                     'NotProcessed': []}}
        if wf_cs['Request'] == '':
          continue
        fileDescendants = {}
        res = uod.execute(prod_id, prod_job_id, wms_job_id,
                          workflowStatus, stepStatus,
                          wf_cs, s_cs,
                          step_number, step_id,
                          SEs=['SomeSE'],
                          fileDescendants=fileDescendants)
        assert res['OK'] is True
#            if transferAndRegisterFileFailover['OK']:
#              self.assertTrue( res['OK'] )
#            else:
#              self.assertFalse( res['OK'] )
        os.remove('foo.txt')
        os.remove('bar.txt')


def test__getLFNsForBKRegistration(mocker):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())

  f1 = File()
  f1.LFN = '/a/1.txt'
  f2 = File()
  f2.LFN = '/a/2.txt'
  f3 = File()
  f3.LFN = '/a/3.txt'

  o1 = Operation()
  o1.Type = 'RegisterFile'
  o1.addFile(f1)
  o2 = Operation()
  o2.Type = 'RegisterFile'
  o2.addFile(f2)
  o3 = Operation()
  o3.Type = 'ForwardDISET'
  o4 = Operation()
  o4.Type = 'RegisterFile'
  o4.addFile(f1)
  o4.addFile(f3)

  r = Request()
  r.addOperation(o4)
  r.addOperation(o1)
  r.addOperation(o2)
  r.addOperation(o3)

  uod = UploadOutputData(bkClient=bkc_mock, dm=dm_mock)
  uod.request = r

  lfns = set(['/a/1.txt', '/a/5.txt', '/a/6.txt', '/a/3.txt'])
  lfnsForBkReg = uod._getLFNsForBKRegistration(lfns)
  assert sorted(lfnsForBkReg) == sorted(['/a/5.txt', '/a/6.txt'])


def test__cleanUp(mocker):
  mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())

  f1 = File()
  f1.LFN = '/a/1.txt'
  f2 = File()
  f2.LFN = '/a/2.txt'
  f3 = File()
  f3.LFN = '/a/3.txt'

  o1 = Operation()
  o1.Type = 'RegisterFile'
  o1.addFile(f1)
  o2 = Operation()
  o2.Type = 'RegisterFile'
  o2.addFile(f2)
  o3 = Operation()
  o3.Type = 'ForwardDISET'
  o4 = Operation()
  o4.Type = 'RegisterFile'
  o4.addFile(f1)
  o4.addFile(f3)

  r = Request()
  r.addOperation(o4)
  r.addOperation(o1)
  r.addOperation(o2)
  r.addOperation(o3)

  uod = UploadOutputData(bkClient=bkc_mock, dm=dm_mock)
  uod.failoverTransfer = ft_mock
  uod.request = r

  expected = Request()
  expected.addOperation(o3)
  removeOp = Operation()
  removeOp.Type = 'RemoveFile'
  fileRemove1 = File()
  fileRemove1.LFN = '/a/1.txt'
  fileRemove2 = File()
  fileRemove2.LFN = '/a/2.txt'
  fileRemove3 = File()
  fileRemove3.LFN = '/a/notPresent.txt'
  removeOp.addFile(fileRemove1)
  removeOp.addFile(fileRemove3)
  removeOp.addFile(fileRemove2)
  expected.addOperation(removeOp)

  uod._cleanUp({'1.txt': {'lfn': '/a/1.txt'},
                '2.txt': {'lfn': '/a/2.txt'},
                'notPresent.txt': {'lfn': '/a/notPresent.txt'}})

  for opsR, opsE in zip(uod.request, expected):
    opsRLoaded = json.loads(str(opsR))
    opsELoaded = json.loads(str(opsE))
    for k in set(opsRLoaded) | set(opsELoaded):
      if k == "Files":
        assert (
            sorted(opsRLoaded[k], key=lambda x: x["LFN"]) ==
            sorted(opsELoaded[k], key=lambda x: x["LFN"])
        )
      else:
        assert opsRLoaded[k] == opsELoaded[k]



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module that contains client access to the transformation DB handler.

This is a very simple extension to the DIRAC one
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient as DIRACTransformationClient
from LHCbDIRAC.ProductionManagementSystem.Utilities.StateMachine import ProductionsStateMachine
from LHCbDIRAC.TransformationSystem.Utilities.StateMachine import TransformationFilesStateMachine


class TransformationClient(DIRACTransformationClient):

  """Exposes the functionality available in the
  LHCbDIRAC/TransformationHandler.

  This inherits the DIRAC base Client for direct execution of server functionality.
  The following methods are available (although not visible here).

  BK query manipulation
      deleteBookkeepingQuery(queryID)
      deleteTransformationBookkeepingQuery(transName)
      addBookkeepingQuery(transID,queryDict)
      getBookkeepingQuery(transName)
  """

  def __init__(self, **kwargs):
    DIRACTransformationClient.__init__(self, **kwargs)
    self.dataProcessingTypes = Operations().getValue('Transformations/DataProcessing', [])

  def addTransformation(self, transName, description, longDescription, transfType, plugin, agentType, fileMask,
                        transformationGroup='General',
                        groupSize=1,
                        inheritedFrom=0,
                        body='',
                        maxTasks=0,
                        eventsPerTask=0,
                        addFiles=True,
                        inputMetaQuery=None,
                        outputMetaQuery=None,
                        bkQuery=None,
                        timeout=1800):
    res = super(TransformationClient, self).addTransformation(transName, description, longDescription,
                                                              transfType, plugin, agentType, fileMask,
                                                              transformationGroup=transformationGroup,
                                                              groupSize=groupSize,
                                                              inheritedFrom=inheritedFrom,
                                                              body=body,
                                                              maxTasks=maxTasks,
                                                              eventsPerTask=eventsPerTask,
                                                              addFiles=addFiles,
                                                              inputMetaQuery=inputMetaQuery,
                                                              outputMetaQuery=outputMetaQuery,
                                                              timeout=timeout)
    if not res['OK']:
      return res
    transID = res['Value']
    if bkQuery:
      res = self._getRPC().addBookkeepingQuery(transID, bkQuery)
      if not res['OK']:
        gLogger.error("Failed to publish BKQuery for transformation", "%s %s" % (transID, res['Message']))
        return res
    return S_OK(transID)

  def _applyTransformationStatusStateMachine(self, transIDAsDict, dictOfProposedstatus, force):
    """Performs a state machine check for productions when asked to change the
    status."""
    originalStatus, transformationType = list(transIDAsDict.values())[0][0:2]
    proposedStatus = list(dictOfProposedstatus.values())[0]
    if force:
      return proposedStatus
    else:
      if transformationType in self.dataProcessingTypes:
        stateChange = ProductionsStateMachine(originalStatus).setState(proposedStatus)
        if not stateChange['OK']:
          return originalStatus
        else:
          return stateChange['Value']
      else:
        return proposedStatus

  def _applyTransformationFilesStateMachine(self, tsFilesAsDict, dictOfProposedLFNsStatus, force):
    """Apply LHCb state machine for transformation files."""
    newStatuses = dict()
    for lfn, newStatus in dictOfProposedLFNsStatus.items():
      if lfn in tsFilesAsDict:
        currentStatus = tsFilesAsDict[lfn][0]
        if force:
          # We do whatever is requested
          newStatus = dictOfProposedLFNsStatus[lfn]
        else:
          # Special case for Assigned -> Unused
          if currentStatus.lower() == 'assigned' and newStatus.lower() == 'unused':
            errorCount = tsFilesAsDict[lfn][1]
            if errorCount and ((errorCount % self.maxResetCounter) == 0):
              newStatus = 'MaxReset'

          tfsm = TransformationFilesStateMachine(currentStatus)
          stateChange = tfsm.setState(newStatus)
          if stateChange['OK']:
            newStatus = stateChange['Value']
        # Only worth setting the status if different from current one
        if newStatus.lower() != currentStatus.lower():
          newStatuses[lfn] = newStatus

    return newStatuses



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Takes an xmlSummary along with its IDs and creates a JSON file containing
all the counters in the xmlSummary."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import xml.etree.ElementTree as ET
from DIRAC import gLogger, S_ERROR, S_OK


def readXMLfile(xmlFile, jobID, prodID, wmsID, jsonFileName='errors_xmlSummary.json'):
  """The main execution function.

  :param str xmlFile: the name of the XML file
  :param str jobID: the job ID of the data
  :param str prodID: the production ID of the data
  :param str wmsID: the wms ID of the data
  """
  res = extractRoot(xmlFile)
  if not res['OK']:
    return res
  root = res['Value']
  res = extractData(root, jobID, prodID, wmsID)
  if not res['OK']:
    return res
  jsonData = res['Value']
  gLogger.notice('Attempting to read %s and put save it as %s' % (xmlFile, jsonFileName))
  createJSONfile(jsonData, jsonFileName)
  return S_OK()

##############################################


def extractRoot(xmlFile):
  """Takes the root element in the XML file.

  :param str xmlFile: the name of the XML file
  """
  try:
    gLogger.notice('Extracting the tree from the XML file %s' % xmlFile)
    tree = ET.parse(xmlFile)
  except Exception as inst:
    gLogger.error('ERROR: Cannot extract tree from XML file %s' % xmlFile)
    return S_ERROR(inst)
  return S_OK(tree.getroot())

##############################################


def extractData(root, jobID, prodID, wmsID):
  """Takes the root of the XML file and the ids of the job, and converts the
  json file from the XML file.

  :param str jobID: the job ID
  :param str prodID: the production ID
  :param str wmsID: the wms ID
  """
  gLogger.notice('Extracting data from the root of the XML file')
  idDict = {}
  idDict['JobID'] = jobID
  idDict['ProductionID'] = prodID
  idDict['wmsID'] = wmsID

  tempDict = {}
  result = {}

  if root.iter('counter') is None:
    return S_ERROR('ERROR: No counters in XML file!')
  for counter in root.iter('counter'):
    try:
      counterValue = int(counter.text)
    except ValueError as inst:
      return S_ERROR(str(inst))
    tempDict[counter.attrib['name']] = counterValue

  # Checks if dict is empty (empty dicts evaluate to False)
  if not tempDict:
    gLogger.warn('WARNING: XML file is empty')

  result['Counters'] = tempDict
  result['Counters']['ID'] = idDict
  return S_OK(json.dumps(result, indent=2))

##############################################


def createJSONfile(jsonData, jsonFileName):
  """Creates a JSON file given a string containing the data.

  :param str jsonData: the string that contains the data
  :param str jsonFileName: the name of the resulting JSON file
  """
  with open(jsonFileName, 'w') as f:
    gLogger.notice('Writing the XML file to JSON file %s' % jsonFileName)
    f.write(jsonData)

# The file is run as following:
# readXMLfile(xmlFileName.xml, jobID, prodID, wmsID, jsonFileName.json)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""reimplementation of the dictionary."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from six.moves import UserDict

from DIRAC import gLogger
from LHCbDIRAC.BookkeepingSystem.Client import IndentMaker


__RCSID__ = "$Id$"


class odict(UserDict):
  """user defined dictionary."""
  #############################################################################

  def __init__(self, dict=None):
    """initialize."""
    self._keys = []
    UserDict.__init__(self, dict)

  #############################################################################
  def __delitem__(self, key):
    """delete."""
    UserDict.__delitem__(self, key)
    self._keys.remove(key)

  #############################################################################
  def __setitem__(self, key, item):
    """set."""
    UserDict.__setitem__(self, key, item)
    if key not in self._keys:
      self._keys.append(key)

  #############################################################################
  def clear(self):
    """clear."""
    UserDict.clear(self)
    self._keys = []

  #############################################################################
  def copy(self):
    """copy."""
    dict = UserDict.copy(self)
    dict._keys = self._keys[:]
    return dict

  #############################################################################
  def items(self):
    """items."""
    return zip(self._keys, self.values())

  #############################################################################
  def keys(self):
    """keys."""
    return self._keys

  #############################################################################
  def popitem(self):
    """popitem."""
    try:
      key = self._keys[-1]
    except IndexError:
      raise KeyError('dictionary is empty')

    val = self[key]
    del self[key]

    return (key, val)

  #############################################################################
  def setdefault(self, key, failobj=None):
    """default value."""
    UserDict.setdefault(self, key, failobj)
    if key not in self._keys:
      self._keys.append(key)

  #############################################################################
  def update(self, dictIn):
    """update."""
    UserDict.update(self, dictIn)
    for key in list(dictIn):
      if key not in self._keys:
        self._keys.append(key)

  #############################################################################
  def values(self):
    """values."""
    return map(self.get, self._keys)


############################################################################
class Entity(dict):
  """Entity class."""
  #############################################################################

  def __init__(self, properties={}):
    """initialize an Entity."""
    # odict.__init__(self)
    if isinstance(properties, list):
      for key in properties:
        self[key] = None  # find a simpler way to declare all keys
    elif isinstance(properties, type(odict)):
      if properties:
        self.update(properties)
    elif isinstance(properties, dict):
      if properties:
        self.update(properties.items())
    else:
      gLogger.warn("Cannot create Entity from properties:" + str(properties))

  #############################################################################
#
#  def __repr__(self):
#    if len(self) == 0 :
#      s = "{\n " + str(None) + "\n}"
#    else:
#      s  = "{"
#      keys = self.keys()
#      for key in keys:
#        #if key == 'fullpath':
#          s += "\n " + str(key) + " : "
#          value = self[key]
#
#          if isinstance(value, dict):
#            value = Entity(value)
#            s += "\n" + IndentMaker.prepend(str(value), (len(str(key))+3)*" ")
#            #childrenString += str(Entity(child)) + "\n"
#          else:
#             s +=  str(value)
#          s += "\n}"
#     #        s = IndentMaker.prepend(s, "_______")
#    return s
#
  def __repr__(self):
    """print."""
    if not self:
      string = "{\n " + str(None) + "\n}"
    else:
      string = "{"
      keys = self.keys()
      if 'fullpath' in keys:
        string += '\n' + 'fullpath: ' + str(self['fullpath'])
      for key in keys:
        if key not in ('name', 'level', 'fullpath', 'expandable', 'selection', 'method', 'showFiles'):
          string += "\n " + str(key) + " : "
          value = self[key]
          # some entities do not have this key. Ignore then.
          try:
            if key in self['not2show']:
              string += '-- not shown --'
              continue
          except Exception:
            pass
          if isinstance(value, dict):
            value = Entity(value)
            string += "\n" + IndentMaker.prepend(str(value), (len(str(key)) + 3) * " ")
          else:
            string += str(value)
      else:
        for key in keys:
          if key not in ('name', 'fullpath', 'FileName'):
            value = self[key]
            if isinstance(value, dict):
              value = Entity(value)
              string += "\n" + IndentMaker.prepend(str(value), (len(str(key)) + 3) * " ")
            else:
              string += str(value)

      string += "\n}"
  #        string = IndentMaker.prepend(string, "_______")
    return string



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.AccountingSystem.private.Plotters.PopularityPlotter.

PopularityPlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter.BaseReporter
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter

from LHCbDIRAC.AccountingSystem.Client.Types.Popularity import Popularity

__RCSID__ = "$Id$"

# FIXME: refactor _reportMethods
# FIXME: refactor _plotMethods


class PopularityPlotter(BaseReporter):
  """PopularityPlotter as extension of BaseReporter."""

  _typeName = "Popularity"
  _typeKeyFields = [dF[0] for dF in Popularity().definitionKeyFields]
  # FIXME: WTF is this ????, here includes StorageElement !!!
  _noSEtypeKeyFields = [dF[0] for dF in Popularity().definitionKeyFields]
  _noSEGrouping = (", ".join("%s" for f in _noSEtypeKeyFields), _noSEtypeKeyFields)

  # .............................................................................
  # data Usage

  _reportDataUsageName = "Data Usage"

  def _reportDataUsage(self, reportRequest):
    """Reports the data usage, from the Accounting DB.

    :param reportRequest: <dict>
      { 'groupingFields' : ( '%s', [ 'EventType' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'EventType' : '90000000' }
      }

    returns S_OK / S_ERROR
      { 'graphDataDict' : { '90000000' : { 1355616000L : 123.456789,
                                           1355702400L : 78.901234500000001 }
                                         },
        'data'          : { '90000000' : { 1355616000L : 123456.789,
                                           1355702400L : 78901.234500000006 }
                                         },
        'unit'          : 'kfiles',
        'granularity'   : 86400
      }
    """

    selectString = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectString + ", %s, %s, SUM(%s)/SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'Usage', 'entriesInBucket'
                                                          ]
                    )

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                PopularityPlotter._noSEGrouping,
                                {'convertToGranularity': 'sum', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxValue = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "files")

    # 3rd value, maxValue is not used
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName
                 })

  def _plotDataUsage(self, reportRequest, plotInfo, filename):
    """Creates <filename>.png file containing information regarding the data
    usage.

    :param reportRequest: <dict>
       { 'grouping'       : 'EventType',
         'groupingFields' : ( '%s', [ 'EventType' ] ),
         'startTime'      : 1355663249.0,
         'endTime'        : 1355749690.0,
         'condDict'       : { 'StorageElement' : 'CERN' }
       }
    :param plotInfo: <dict> ( output of _reportDataUsage )
       { 'graphDataDict' : { '90000001' : { 1355616000L : 223.45678899999999,
                                            1355702400L : 148.90123449999999 },
                             '90000000' : { 1355616000L : 123.456789,
                                            1355702400L : 78.901234500000001 }
                            },
         'data'          : { '90000001' : { 1355616000L : 223456.78899999999,
                                            1355702400L : 148901.23449999999 },
                             '90000000' : { 1355616000L : 123456.789,
                                            1355702400L : 78901.234500000006 }
                            },
         'unit'          : 'kfiles',
         'granularity': 86400
        }
    :param filename: <str>
      '_plotDataUsage'

    returns S_OK / S_ERROR
       { 'plot': True, 'thumbnail': False }
    """

    startEpoch = reportRequest['startTime']
    endEpoch = reportRequest['endTime']
    granularity = plotInfo['granularity']
    dataDict = plotInfo['graphDataDict']

    metadata = {
        'title': "Data Usage grouped by %s" % reportRequest['grouping'],
        'starttime': startEpoch,
                 'endtime': endEpoch,
                 'span': granularity,
                 'ylabel': plotInfo['unit']
    }

    dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  # .............................................................................
  # normalized data Usage

  _reportNormalizedDataUsageName = "Normalized Data Usage"

  def _reportNormalizedDataUsage(self, reportRequest):
    """Reports the normalized data usage, from the Accounting DB.

    :param reportRequest: <dict>
      { 'groupingFields' : ( '%s', [ 'EventType' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'StorageElement' : 'CERN' }
      }

    returns S_OK / S_ERROR
      { 'graphDataDict' : { '90000001' : { 1355616000L : 223.45678899999999,
                                           1355702400L : 148.90123449999999 },
                            '90000000' : { 1355616000L : 123.456789,
                                           1355702400L : 78.901234500000001 }
                          },
        'data'          : { '90000001' : { 1355616000L : 223456.78899999999,
                                           1355702400L : 148901.23449999999 },
                            '90000000' : { 1355616000L : 123456.789,
                                           1355702400L : 78901.234500000006 }
                          },
        'unit'          : 'kfiles',
        'granularity'   : 86400
      }
    """

    selectString = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectString + ", %s, %s, SUM(%s)/SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'NormalizedUsage', 'entriesInBucket'
                                                          ]
                    )
    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                PopularityPlotter._noSEGrouping,
                                {'convertToGranularity': 'sum', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxValue = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "files")

    # 3rd value, maxValue is not used
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName
                 })

  def _plotNormalizedDataUsage(self, reportRequest, plotInfo, filename):
    """Creates <filename>.png file containing information regarding the
    normalized data usage.

    :param reportRequest: <dict>
       { 'grouping'       : 'EventType',
         'groupingFields' : ( '%s', [ 'EventType' ] ),
         'startTime'      : 1355663249.0,
         'endTime'        : 1355749690.0,
         'condDict'       : { 'StorageElement' : 'CERN' }
       }
    :param plotInfo: <dict> ( output of _reportNormalizedDataUsage )
       { 'graphDataDict' : { '90000001' : { 1355616000L : 223.45678899999999,
                                            1355702400L : 148.90123449999999 },
                             '90000000' : { 1355616000L : 123.456789,
                                            1355702400L : 78.901234500000001 }
                            },
         'data'          : { '90000001' : { 1355616000L : 223456.78899999999,
                                            1355702400L : 148901.23449999999 },
                             '90000000' : { 1355616000L : 123456.789,
                                            1355702400L : 78901.234500000006 }
                            },
         'unit'          : 'kfiles',
         'granularity': 86400
        }
    :param filename: <str>
      '_plotNormalizedDataUsage'

    returns S_OK / S_ERROR
       { 'plot': True, 'thumbnail': False }
    """

    startEpoch = reportRequest['startTime']
    endEpoch = reportRequest['endTime']
    dataDict = plotInfo['graphDataDict']
    granularity = plotInfo['granularity']

    metadata = {
        'title': "Normalized Data Usage grouped by %s" % reportRequest['grouping'],
        'starttime': startEpoch,
                 'endtime': endEpoch,
                 'span': granularity,
                 'ylabel': plotInfo['unit']
    }

    dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve files for a given run."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ... Run' % Script.scriptName,
      'Arguments:',
      '  Run:      Run number (integer)']))
  Script.parseCommandLine(ignoreErrors=True)
  args = Script.getPositionalArgs()

  try:
    runID = int(args[0])
  except (ValueError, IndexError):
    Script.showHelp(exitCode=1)

  exitCode = 0

  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  res = BookkeepingClient().getRunFiles(runID)
  if not res['OK']:
    print('Failed to retrieve run files: %s' % res['Message'])
    exitCode = 2
  else:
    if not res['Value']:
      print('No files found for run %s' % runID)
    else:
      print('%s %s %s %s' % ('FileName'.ljust(100), 'Size'.ljust(10), 'GUID'.ljust(40), 'Replica'.ljust(8)))
      for lfn in sorted(res['Value']):
        size = res['Value'][lfn]['FileSize']
        guid = res['Value'][lfn]['GUID']
        hasReplica = res['Value'][lfn]['GotReplica']
        print('%s %s %s %s' % (lfn.ljust(100), str(size).ljust(10), guid.ljust(40), str(hasReplica).ljust(8)))

  DIRAC.exit(exitCode)


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""For a given LFN directory, check the files that are registered in the FC and
checks that they exist on the SE, and in Bookkeeping, with the  correct
ReplicaFlag."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script
  unit = 'TB'
  sites = []
  dir = ''
  fileType = ''
  prods = []
  prodID = ''
  verbose = False
  Script.registerSwitch("u:", "Unit=", "   Unit to use [%s] (MB,GB,TB,PB)" % unit)
  Script.registerSwitch("D:", "Dir=", "  directory to be checked: mandatory argument")
  Script.registerSwitch("f:", "Output=", " output file name [dirac-dms-chec-dir-cont.out]")
  Script.registerSwitch("v", "Verbose", " use this option for verbose output [False]")

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ...' %
      Script.scriptName, ]))

  Script.parseCommandLine(ignoreErrors=False)

  from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
  from DIRAC.Resources.Storage.StorageElement import StorageElement
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  bkClient = BookkeepingClient()

  outputFileName = 'dirac-dms-chec-dir-cont.out'

  for switch in Script.getUnprocessedSwitches():
    if switch[0].lower() == "u" or switch[0].lower() == "unit":
      unit = switch[1]
    if switch[0] == "D" or switch[0].lower() == "dir":
      dir = switch[1]
    if switch[0] == "f" or switch[0].lower() == "output":
      outputFile = switch[1]
      outputFileName = outputFile
    if switch[0] == "v" or switch[0].lower() == "verbose":
      verbose = True

  if verbose:
    print('Verbose output')

  if not dir:
    print('One directory should be provided!')
    Script.showHelp()

  scaleDict = {
      'MB': 1000 * 1000.0,
      'GB': 1000 * 1000 * 1000.0,
      'TB': 1000 * 1000 * 1000 * 1000.0,
      'PB': 1000 * 1000 * 1000 * 1000 * 1000.0,
  }
  if unit not in scaleDict:
    Script.showHelp()
  scaleFactor = scaleDict[unit]

  currentDir = dir
  print('Obtaining the catalog contents for %s directory' % currentDir)
  res = FileCatalog().listDirectory(currentDir)
  if not res['OK']:
    print('ERROR: Cannot get directory content')
    DIRAC.exit(-1)

  successfulDirs = res['Value']['Successful']
  failedDirs = res['Value']['Failed']
  print('Failed directories: % s' % list(failedDirs))
  print('Successful directories: % s' % list(successfulDirs))

  if not successfulDirs:
    print('No directory to analyse. Exit.')
    DIRAC.exit(0)

  if verbose:
    print('Analysing directory: %s ' % currentDir)
  dirData = successfulDirs[currentDir]
  NumOfFilesInLFC = len(dirData['Files'])
  if verbose:
    print('Number of files registered in LFC: %d ' % NumOfFilesInLFC)
  LFNsInLFC = list(dirData['Files'])
  # print 'List of lfns in lfc: ' , LFNsInLFC

  allFiles = {}
  dirContents = res['Value']['Successful'][currentDir]
  allFiles.update(dirContents['Files'])

  fp = open(outputFileName, "w")

  zeroReplicaFiles = []
  zeroSizeFiles = []
  allReplicaDict = {}
  allMetadataDict = {}
  problematicFiles = {}
  replicasPerSE = {}
  n = 10
  totalSoFar = 0
  fp.write("-------- Checks FC -> SE ---------------------------------------------------------\n")
  for lfn, lfnDict in allFiles.items():
    # checks LFC -> SE
    totalSoFar += 1
    if totalSoFar % n == 0:
      print('%d LFNs processed so far. %d left' % (totalSoFar, len(LFNsInLFC) - totalSoFar))
    if verbose:
      fp.write("LFN: %s\n" % lfn)
    lfnReplicas = []
    for se, replicaDict in lfnDict['Replicas'].items():
      # print 'SE: %s -- replica: %s ' % ( se, replicaDict )
      lfnReplicas.append(se)
      if not lfnReplicas:
        zeroReplicaFiles.append(lfn)
    allReplicaDict[lfn] = lfnReplicas
    allMetadataDict[lfn] = lfnDict['MetaData']
    if lfnDict['MetaData']['Size'] == 0:
      zeroSizeFiles.append(lfn)
    if verbose:
      fp.write("All replicas: %s \n" % lfnReplicas)
    # check each replica if it is exists on the storage
    for se in lfnReplicas:
      if se not in replicasPerSE:
        replicasPerSE[se] = []
      replicasPerSE[se].append(lfn)
      if verbose:
        fp.write("Checking on storage LFN, SE: %s %s\n" % (lfn, se))
      res = StorageElement(se).getFileMetadata(lfn)
      if not res['OK']:
        fp.write("ERROR: could not get storage file metadata! %s - %s \n" % (lfn, se))
        if lfn not in problematicFiles:
          problematicFiles[lfn] = {}
        if 'BadReplicas' not in problematicFiles[lfn]:
          problematicFiles[lfn]['BadReplicas'] = []
        problematicFiles[lfn]['BadReplicas'].append(se)
        continue
      if lfn in res['Value']['Failed']:
        fp.write("ERROR: bad LFN! %s\n" % lfn)
        if lfn not in problematicFiles:
          problematicFiles[lfn] = {}
          if 'BadPFN' not in problematicFiles[lfn]:
            problematicFiles[lfn]['BadPFN'] = []
          problematicFiles[lfn]['BadPFN'].append(se)
      elif lfn in res['Value']['Successful']:
        if verbose:
          fp.write("Replica is ok\n")
    fp.flush()

  BkkChecks = False
  print("-------- Checks LFC -> Bkk ---------------------------------------------------------")
  fp.write("-------- Checks LFC -> Bkk ---------------------------------------------------------\n")
  res = bkClient.getFileMetadata(LFNsInLFC)
  if res['OK']:
    BkkChecks = True
    metadata = res['Value']['Successful']
    missingLFNs = [lfn for lfn in LFNsInLFC if metadata.get(lfn, {}).get('GotReplica') is None]
    noFlagLFNs = [lfn for lfn in LFNsInLFC if metadata.get(lfn, {}).get('GotReplica') == 'No']
    okLFNs = [lfn for lfn in LFNsInLFC if metadata.get(lfn, {}).get('GotReplica') == 'Yes']
    if verbose:
      print(
          "Out of %d files, %d have a replica flag in the BK, %d are not in the BK and %d don't have the flag" %
          (len(LFNsInLFC), len(okLFNs), len(missingLFNs), len(noFlagLFNs))
      )

  fp.write(" ++++++++++++++++++++++++++++++ Final summary ++++++++++++++++++++++++++\n")
  fp.write(" +++++++++++++++++++++++++++++ Checks LFC -> SE: ++++++++++++++++++++++++++++\n")
  fp.write("Replicas per SE:\n")
  for se in replicasPerSE:
    fp.write("SE: %s has %d replicas\n" % (se, len(replicasPerSE[se])))
    if verbose:
      for r in replicasPerSE[se]:
        fp.write("%s\n" % r)

  directoryConsistency = True
  if zeroReplicaFiles:
    fp.write("Files with zero replicas: %s \n" % zeroReplicaFiles)
    directoryConsistency = False
  if zeroSizeFiles:
    fp.write("Files with zero size: %s\n" % zeroSizeFiles)
    directoryConsistency = False
  if problematicFiles:
    fp.write("Found some problematic replicas: \n")
    for lfn in problematicFiles:
      fp.write("LFN: %s\n" % lfn)
      for k in problematicFiles[lfn]:
        fp.write("%s : %s \n" % (k, problematicFiles[lfn][k]))
    directoryConsistency = False
  fp.write(" +++++++++++++++++++++++++++++++ Checks LFC -> Bookkeeping: +++++++++++++++++++++++++++++\n")

  if BkkChecks:
    fp.write("Out of %d files:\n %d are OK - ReplicaFlag=Yes\n %d are not in the BK\n %d have ReplicaFlag=No\n" %
             (len(LFNsInLFC), len(okLFNs), len(missingLFNs), len(noFlagLFNs)))
    if len(missingLFNs) > 0:
      directoryConsistency = False
      fp.write("LFNs present in LFC but missing from Bookkeeping:\n")
      for lfn in missingLFNs:
        fp.write("%s\n" % lfn)
    if len(noFlagLFNs) > 0:
      directoryConsistency = False
      fp.write("LFNs present in LFC but with ReplicaFlag=No in Bookkeeping:\n")
      for lfn in noFlagLFNs:
        fp.write("%s\n" % lfn)
  else:
    fp.write("Bookkeeping didn't return any valid result for the LFNs\n")

  if directoryConsistency:
    fp.write("--------->All replicas in LFC have been checked on Storage. Ok!!\n")
  else:
    fp.write("--------->Directory is not consistent\n")

  fp.close()
  print("Summary written to file: %s" % outputFileName)
  DIRAC.exit(0)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for Workflow Modules."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

import unittest
import os
import copy
import shutil
from textwrap import dedent

from mock import MagicMock, patch

from DIRAC import gLogger
from DIRAC.TransformationSystem.Client.FileReport import FileReport
# mocks
from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from DIRAC.Resources.Catalog.test.mock_FC import fc_mock


from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary
# mocks
from LHCbDIRAC.Workflow.Modules.mock_Commons import step_commons, wf_commons, \
    prod_id, prod_job_id, wms_job_id, \
    workflowStatus, stepStatus, step_id, step_number,\
    rc_mock
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock

# sut
from LHCbDIRAC.Workflow.Modules.AnalyseXMLSummary import AnalyseXMLSummary


__RCSID__ = "$Id$"


class ModulesTestCase(unittest.TestCase):
  """Base class for the Modules test cases."""

  def setUp(self):

    gLogger.setLevel('DEBUG')
    self.maxDiff = None

    self.jsu_mock = MagicMock()
    self.jsu_mock.setJobApplicationStatus.return_value = {'OK': True, 'Value': ''}

    self.jsu_mock = MagicMock()
    self.jsu_mock.setJobApplicationStatus.return_value = {'OK': True, 'Value': ''}

    self.ft_mock = MagicMock()
    self.ft_mock.transferAndRegisterFile.return_value = {'OK': True, 'Value': {'uploadedSE': ''}}
    self.ft_mock.transferAndRegisterFileFailover.return_value = {'OK': True, 'Value': {}}
    self.ft_mock.request = rc_mock
    self.ft_mock.FileCatalog = fc_mock

    self.nc_mock = MagicMock()
    self.nc_mock.sendMail.return_value = {'OK': True, 'Value': ''}

    self.xf_o_mock = MagicMock()
    self.xf_o_mock.inputFileStats = {'a': 1, 'b': 2}
    self.xf_o_mock.outputFileStats = {'a': 1, 'b': 2}
    self.xf_o_mock.analyse.return_value = True

    self.jobStep_mock = MagicMock()
    self.jobStep_mock.commit.return_value = {'OK': True, 'Value': ''}
    self.jobStep_mock.setValuesFromDict.return_value = {'OK': True, 'Value': ''}
    self.jobStep_mock.checkValues.return_value = {'OK': True, 'Value': ''}

  def tearDown(self):
    for fileProd in [
        'appLog',
        'foo.txt',
        'aaa.Bhadron.dst',
        'bbb.Calibration.dst',
        'bar.py',
        'aLongLog.log',
        'bookkeeping_123_00000456_321.xml',
        'aLongLog.log.gz',
        'ccc.charm.mdst',
        'ccc.charm.mdst',
        'prova.txt',
        'aLog.log',
        'BAR.txt',
        'FooBAR.ext.txt',
        'foo_1.txt',
        'bar_2.py',
        'bar.txt',
        'ErrorLogging_Step1_coredump.log',
        '123_00000456_request.xml',
        'lfn1',
        'lfn2',
        'XMLSummaryFile',
        'aaa.bhadron.dst',
        'bbb.calibration.dst',
        'ProductionOutputData',
        'data.py',
        '123_00000456_request.json',
        '00000123_00000456.tar',
        'someOtherDir',
        'DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK',
        'myfoo.blah',
        'prodConf_someApp__.py',
            'prodConf_someApp___.py']:
      try:
        os.remove(fileProd)
      except OSError:
        continue

    for directory in ['./job', 'job']:
      try:
        shutil.rmtree(directory)
      except BaseException:
        continue


@patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
class AnalyseXMLSummarySuccess(ModulesTestCase):

  #################################################

  def test_execute(self, _patch):

    axlf = AnalyseXMLSummary(bkClient=bkc_mock, dm=dm_mock)
    axlf.stepInputData = ['some.sdst', '00012345_00006789_1.sdst']
    axlf.jobType = 'merge'

    logAnalyser = MagicMock()

    logAnalyser.return_value = True
    axlf.logAnalyser = logAnalyser
    axlf.XMLSummary_o = self.xf_o_mock
    axlf.nc = self.nc_mock
    axlf.XMLSummary = 'XMLSummaryFile'
    with open(axlf.XMLSummary, 'w') as f:
      f.write(dedent(
          """<?xml version="1.0" encoding="UTF-8"?>

          <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                  <success>True</success>
                  <step>finalize</step>
                  <usage>
                          <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                  </usage>
                  <input>
                          <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim" status="full">200</file>
                  </input>
                  <output>
                          <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi" status="full">200</file>
                  </output>
          </summary>
          """  # noqa
      ))

    # no errors, all ok
    for wf_cs in copy.deepcopy(wf_commons):
      for s_cs in step_commons:
        self.assertTrue(axlf.execute(prod_id, prod_job_id, wms_job_id,
                                     workflowStatus, stepStatus,
                                     wf_cs, s_cs,
                                     step_number, step_id)['OK'])

    # logAnalyser gives errors
    axlf.jobType = 'reco'

    logAnalyser.return_value = False
    axlf.logAnalyser = logAnalyser

    for wf_cs in copy.deepcopy(wf_commons):
      for s_cs in step_commons:
        self.assertTrue(axlf.execute(prod_id, prod_job_id, wms_job_id,
                                     workflowStatus, stepStatus,
                                     wf_cs, s_cs,
                                     step_number, step_id)['OK'])

  def test__basicSuccess(self, _patch):

    axlf = AnalyseXMLSummary(bkClient=bkc_mock, dm=dm_mock)

    with open('XMLSummaryFile', 'w') as f:
      f.write(dedent(
          """<?xml version="1.0" encoding="UTF-8"?>

          <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                  <success>True</success>
                  <step>finalize</step>
                  <usage>
                          <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                  </usage>
                  <input>
                          <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim" status="full">200</file>
                  </input>
                  <output>
                          <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi" status="full">200</file>
                  </output>
          </summary>
          """  # noqa
      ))
    axlf.XMLSummary_o = XMLSummary('XMLSummaryFile')
    res = axlf._basicSuccess()
    self.assertFalse(res)

    axlf.XMLSummary_o.inputFileStats = {'full': 2, 'part': 1, 'fail': 0, 'other': 0}
    axlf.XMLSummary_o.inputStatus = [('aa/1.txt', 'full'), ('aa/2.txt', 'part')]
    axlf.inputDataList = ['aa/1.txt', 'aa/2.txt']
    axlf.numberOfEvents = -1
    axlf.fileReport = FileReport()
    axlf.production_id = '123'
    res = axlf._basicSuccess()
    self.assertTrue(res)
    self.assertEqual(axlf.fileReport.statusDict, {'aa/2.txt': 'Problematic'})

    axlf.XMLSummary_o.inputFileStats = {'full': 2, 'part': 0, 'fail': 1, 'other': 0}
    axlf.XMLSummary_o.inputStatus = [('aa/1.txt', 'fail'), ('aa/2.txt', 'full')]
    axlf.inputDataList = ['aa/1.txt', 'aa/2.txt']
    axlf.numberOfEvents = -1
    axlf.fileReport = FileReport()
    axlf.production_id = '123'
    res = axlf._basicSuccess()
    self.assertTrue(res)
    self.assertEqual(axlf.fileReport.statusDict, {'aa/1.txt': 'Problematic'})

    axlf.XMLSummary_o.inputFileStats = {'full': 2, 'part': 0, 'fail': 1, 'other': 0}
    axlf.XMLSummary_o.inputStatus = [('aa/1.txt', 'fail'), ('aa/2.txt', 'full')]
    axlf.inputDataList = ['aa/3.txt']
    axlf.numberOfEvents = -1
    axlf.fileReport = FileReport()
    axlf.production_id = '123'
    res = axlf._basicSuccess()
    self.assertFalse(res)
    self.assertEqual(axlf.fileReport.statusDict, {})

    axlf.XMLSummary_o.inputFileStats = {'full': 2, 'part': 1, 'fail': 1, 'other': 0}
    axlf.XMLSummary_o.inputStatus = [('aa/1.txt', 'fail'), ('aa/2.txt', 'part')]
    axlf.inputDataList = ['aa/1.txt', 'aa/2.txt']
    axlf.numberOfEvents = -1
    axlf.fileReport = FileReport()
    axlf.production_id = '123'
    res = axlf._basicSuccess()
    self.assertTrue(res)
    self.assertEqual(axlf.fileReport.statusDict, {'aa/1.txt': 'Problematic', 'aa/2.txt': 'Problematic'})

    axlf.XMLSummary_o.inputFileStats = {'full': 2, 'part': 1, 'fail': 1, 'other': 0}
    axlf.XMLSummary_o.inputStatus = [('aa/1.txt', 'fail'), ('aa/2.txt', 'part')]
    axlf.inputDataList = ['aa/1.txt', 'aa/2.txt']
    axlf.numberOfEvents = '10'
    axlf.fileReport = FileReport()
    axlf.production_id = '123'
    res = axlf._basicSuccess()
    self.assertTrue(res)
    self.assertEqual(axlf.fileReport.statusDict, {'aa/1.txt': 'Problematic'})


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve information from the Bookkeeping for a given production."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ... ProdID' % Script.scriptName,
      'Arguments:',
      '  ProdID:   Production ID (mandatory)']))
  Script.parseCommandLine(ignoreErrors=True)
  args = Script.getPositionalArgs()

  if len(args) < 1:
    Script.showHelp(exitCode=1)

  exitCode = 0

  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  bk = BookkeepingClient()
  prod = int(args[0])

  res = bk.getProductionInformation(prod)
  if not res['OK']:
    print("ERROR %s: %s" % (str(prod), res['Message']))
    DIRAC.exit(2)

  val = res['Value']
  print("Production Info: ")

  infs = val["Production information"]
  if infs is not None:
    for inf in infs:
      if inf[2] is not None:
        print('    Configuration Name:', inf[0])
        print('    Configuration Version:', inf[1])
        print('    Event type:', inf[2])

  steps = val['Steps']

  for step in steps:
    print("-----------------------")
    print(" StepName: %s " % (step[0]))
    print("    StepId             : %s" % (step[7]))
    print("    ApplicationName    : %s" % (step[1]))
    print("    ApplicationVersion : %s" % (step[2]))
    print("    OptionFiles        : %s" % (step[3]))
    print("    DDB                : %s" % (step[4]))
    print("    CONDDB             : %s" % (step[5]))
    print("    ExtraPackages      : %s" % (step[6]))
    print("    Visible            : %s" % (step[8]))
    print("-----------------------")
  print("Number of Steps  ", val["Number of jobs"][0][0])
  files = val["Number of files"]
  if files:
    print("Total number of files:", files[0][2])
  else:
    print("Total number of files: 0")
  for file in files:
    print("         " + str(file[1]) + ":" + str(file[0]))
  nbevent = val["Number of events"]
  if nbevent:
    print("Number of events")
    print("File Type".ljust(20) + "Number of events".ljust(20) + "Event Type".ljust(20) + "EventInputStat")
    for i in nbevent:
      print(str(i[0]).ljust(20) + str(i[1]).ljust(20) + str(i[2]).ljust(20) + str(i[3]))
  else:
    print("Number of events", 0)
  print('Path: ', val['Path'])


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""  The MC Simulation Template creates workflows for the following simulation
     use-cases:
      WORKFLOW1: Simulation+Selection+Merge
      WORKFLOW2: Simulation+Selection+MCMerge
      WORKFLOW3: Simulation+Selection
      WORKFLOW4: Simulation+MCMerge
      WORKFLOW5: Simulation

    Exotic things you might want to do:
    * run a local test:
      - of the MC: just set the localTestFlag to True
      - of the merging/stripping: set pr.prodsToLaunch to, e.g., [2], and adjust the pr.inputs at the end of the script
    * run only part of the request on the Grid:
      - for the MC: just set pr.prodsToLaunch = [1]
      - for the merge and/or stripping: set pr.prodsToLaunch, then set pr.previousProdID
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from six.moves import range

import ast

from DIRAC.Core.Base import Script
Script.parseCommandLine()

from DIRAC import gConfig, gLogger, exit as DIRACexit
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import ProductionRequest

__RCSID__ = "$Id$"

gLogger = gLogger.getSubLogger('MCSimulation_run.py')
currentSetup = gConfig.getValue('DIRAC/Setup')

pr = ProductionRequest()

stepsList = ['{{p1Step}}']
stepsList.append('{{p2Step}}')
stepsList.append('{{p3Step}}')
stepsList.append('{{p4Step}}')
stepsList.append('{{p5Step}}')
stepsList.append('{{p6Step}}')
stepsList.append('{{p7Step}}')
stepsList.append('{{p8Step}}')
stepsList.append('{{p9Step}}')
stepsList.append('{{p10Step}}')
stepsList.append('{{p11Step}}')
stepsList.append('{{p12Step}}')
stepsList.append('{{p13Step}}')
stepsList.append('{{p14Step}}')
stepsList.append('{{p15Step}}')
stepsList.append('{{p16Step}}')
stepsList.append('{{p17Step}}')
stepsList.append('{{p18Step}}')
stepsList.append('{{p19Step}}')
stepsList.append('{{p20Step}}')
pr.stepsList = stepsList

###########################################
# Configurable and fixed parameters
###########################################

pr.appendName = '{{WorkflowAppendName#GENERAL: Workflow string to append to production name#1}}'

w = '{{w#----->WORKFLOW: choose one below#}}'
w1 = '{{w1#-WORKFLOW1: Simulation#False}}'
w2 = '{{w2#-WORKFLOW2: Simulation(Gauss) + AllOthers#False}}'
w3 = '{{w3#-WORKFLOW3: Simulation(Gauss) + AllOthers + Merge#False}}'

localTestFlag = '{{localTestFlag#GENERAL: Set True for local test#False}}'
validationFlag = '{{validationFlag#GENERAL: Set True for validation prod - will create histograms#False}}'

pr.configName = '{{BKConfigName#GENERAL: BK configuration name e.g. MC #MC}}'
extraOptions = '{{extraOptions#GENERAL: extra options as python dict stepID:options#}}'

targets = '{{Target#PROD-1:MC: Target for MC (e.g. Tier2, ALL, LCG.CERN.cern or BAN:site1:site2#ALL}}'
eventsPerJob = '{{eventsPerJob#PROD-1:MC: Number of events per job#-1}}'
MCPriority = '{{MCPriority#PROD-1:MC: Production priority#0}}'
MCmulticoreFlag = '{{MCMulticoreFLag#PROD-1: multicore flag#True}}'
MCSimulationType = '{{MCSimulationType#PROD-1:MC: type of MCSimulation#MCSimulation}}'
simulationOutputVisFlagSpecial = {}

selectionPlugin = '{{selectionPlugin#PROD-2:Selection: plugin e.g. Standard, BySize#BySize}}'
selectionGroupSize = '{{selectionGroupSize#PROD-2:Selection: input files total size (we\'ll download)#5}}'
selectionPriority = '{{selectionPriority#PROD-2:Selection: Job Priority e.g. 8 by default#8}}'
selectionCPU = '{{selectionCPU#PROD-2:Selection: Max CPU time in secs#100000}}'
removeInputSelection = '{{removeInputSelection#PROD-2:Selection: remove inputs#True}}'
selmulticoreFlag = '{{selMulticoreFLag#PROD-2:Selection: multicore flag#True}}'
selectionOutputVisFlagSpecial = {}

mergingPlugin = '{{MergingPlugin#PROD-3:Merging: plugin e.g. Standard, BySize#BySize}}'
mergingGroupSize = '{{MergingGroupSize#PROD-3:Merging: Group Size e.g. BySize = GB file size#5}}'
mergingPriority = '{{MergingPriority#PROD-3:Merging: Job Priority e.g. 8 by default#8}}'
mergingCPU = '{{mergingCPU#PROD-3:Merging: Max CPU time in secs#100000}}'
removeInputMerge = '{{removeInputMerge#PROD-3:Merging: remove inputs#True}}'
mergemulticoreFlag = '{{mergeMulticoreFLag#PROD-3:Merging: multicore flag#True}}'
mergeOutputVisFlagSpecial = {}

pr.configVersion = '{{mcConfigVersion}}'
pr.eventType = '{{eventType}}'
# Often MC requests are defined with many subrequests but we want to retain
# the parent ID for viewing on the production monitoring page. If a parent
# request is defined then this is used.
pr.parentRequestID = '{{_parent}}'
pr.requestID = '{{ID}}'

if extraOptions:
  pr.extraOptions = ast.literal_eval(extraOptions)
pr.prodGroup = '{{pDsc}}'
pr.dataTakingConditions = '{{simDesc}}'

MCPriority = int(MCPriority)
selectionPriority = int(selectionPriority)
mergingPriority = int(mergingPriority)

removeInputMerge = ast.literal_eval(removeInputMerge)
removeInputSelection = ast.literal_eval(removeInputSelection)

###########################################
# LHCb conventions implied by the above
###########################################

localTestFlag = ast.literal_eval(localTestFlag)
validationFlag = ast.literal_eval(validationFlag)

if localTestFlag:
  pr.testFlag = True
  pr.publishFlag = False
  pr.prodsToLaunch = [1]

pr.outConfigName = pr.configName

w1 = ast.literal_eval(w1)
w2 = ast.literal_eval(w2)
w3 = ast.literal_eval(w3)

if not w1 and not w2 and not w3:
  gLogger.error('Vladimir, I told you to select at least one workflow!')
  DIRACexit(2)

elif w1:
  pr.prodsTypeList = [MCSimulationType]
  pr.outputSEs = ['Tier1_MC-DST']

  pr.stepsInProds = [list(range(1, len(pr.stepsList) + 1))]
  pr.removeInputsFlags = [False]
  pr.priorities = [MCPriority]
  pr.cpus = [100000]
  pr.outputFileSteps = [str(len(pr.stepsList))]
  pr.targets = [targets]
  pr.events = [eventsPerJob]
  pr.groupSizes = [1]
  pr.plugins = ['']
  pr.inputDataPolicies = ['']
  pr.bkQueries = ['']
  pr.multicore = [MCmulticoreFlag]


elif w2:
  pr.prodsTypeList = [MCSimulationType, 'MCReconstruction']
  pr.outputSEs = ['Tier1-Buffer', 'Tier1_MC-DST']

  pr.stepsInProds = [[1, ], list(range(2, len(pr.stepsList) + 1))]
  pr.outputFileSteps = [str(len(pr.stepsInProds[0])),
                        str(len(pr.stepsInProds[1]))]

  pr.removeInputsFlags = [False, removeInputSelection]
  pr.priorities = [MCPriority, selectionPriority]
  pr.cpus = [100000, selectionCPU]
  pr.targets = [targets, '']
  pr.events = [eventsPerJob, -1]
  pr.groupSizes = [1, selectionGroupSize]
  pr.plugins = ['', selectionPlugin]
  pr.inputDataPolicies = ['', 'download']
  pr.bkQueries = ['', 'fromPreviousProd']
  pr.multicore = [MCmulticoreFlag, selmulticoreFlag]


elif w3:
  pr.prodsTypeList = [MCSimulationType, 'MCReconstruction', 'MCMerge']
  pr.outputSEs = ['Tier1-Buffer', 'Tier1-Buffer', 'Tier1_MC-DST']

  pr.stepsInProds = [[1, ], list(range(2, len(pr.stepsList))), [len(pr.stepsList)]]
  pr.outputFileSteps = ['1', str(len(pr.stepsInProds[1])), '1']

  pr.removeInputsFlags = [False, removeInputSelection, removeInputMerge]
  pr.priorities = [MCPriority, selectionPriority, mergingPriority]
  pr.cpus = [100000, selectionCPU, mergingCPU]
  pr.targets = [targets, '', '']
  pr.events = [eventsPerJob, -1, -1]
  pr.groupSizes = [1, selectionGroupSize, mergingGroupSize]
  pr.plugins = ['', selectionPlugin, mergingPlugin]
  pr.inputDataPolicies = ['', 'download', 'download']
  pr.bkQueries = ['', 'fromPreviousProd', 'fromPreviousProd']
  pr.multicore = [MCmulticoreFlag, selmulticoreFlag, mergemulticoreFlag]


numberOfSteps = len(pr.stepsList)
# pr.compressionLvl = [lowCompressionLvl] * numberOfSteps
# pr.compressionLvl[-1] = highCompressionLvl
pr.compressionLvl = ["LOW"] * numberOfSteps
pr.compressionLvl[-1] = "HIGH"

vis = dict([str(i), "N"] for i in range(1, numberOfSteps))
vis[str(numberOfSteps)] = "Y"

pr.outputVisFlag = [vis]
pr.specialOutputVisFlag = [{}]


# In case we want just to test, we publish in the certification/test part of the BKK
if currentSetup == 'LHCb-Certification' or pr.testFlag:
  pr.outConfigName = 'certification'
  pr.configVersion = 'test'

if pr.testFlag:
  pr.extend = '10'
  mergingGroupSize = '1'
  MCCpu = '50000'
  pr.previousProdID = 0  # set this for, e.g., launching only merging

# Validation implies few things, like saving all the outputs, and adding a GAUSSHIST
if validationFlag:
  pr.resolveSteps()
  pr.outConfigName = 'validation'
  # Adding GAUSSHIST to the list of outputs to be produced (by the first step, which is Gauss)
  if 'GAUSSHIST' not in pr.stepsListDict[0]['fileTypesOut']:
    pr.stepsListDict[0]['fileTypesOut'].append('GAUSSHIST')
  pr.outputFileSteps = [''] * len(pr.prodsTypeList)


res = pr.buildAndLaunchRequest()
if not res['OK']:
  gLogger.error("Errors with submission: %s" % res['Message'])
  DIRACexit(2)
else:
  gLogger.always("Submitted %s" % str(res['Value']))



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Bookkeeping utilities."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


from DIRAC import gLogger, S_ERROR

from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient


_IGNORE_PARAMETERS = ['ReplicaFlag', 'Visible', 'MethodName']

# The following parameters can not used to build the query, it requires at least one more parameter.
_ONE = ['FileType', 'ProcessingPass', 'EventType',
        'DataQuality', 'ConfigName', 'ConfigVersion', 'ConditionDescription']

# Two parameter in the list not enough to build the query.
_TWO = ['ConfigName', 'ConfigVersion', 'ConditionDescription',
        'EventType', 'ProcessingPass', 'FileType', 'DataQuality']


def enoughParams(in_dict):
  """Dirty method to check the query parameters and make sure the queries have
  enough parameters."""
  checkingDict = in_dict.copy()
  if not checkingDict:
    return False

  for param in _IGNORE_PARAMETERS:
    if param in checkingDict:
      checkingDict.pop(param)

  if not checkingDict:
    return False

  if len(checkingDict) == 1:
    if not set(checkingDict) - set(_ONE):
      return False

  if len(checkingDict) == 2:
    if not set(checkingDict) - set(_TWO):
      return False
  return True


def checkEnoughBKArguments(func):
  """The decorator used to check the parameters of a given dictionary
  (BkQuery)."""

  def checkMethodArguments(self, *args, **kwargs):
    """This is used to check the conditions of a given query.

    We assume a dictionary can not be empty and it has more than one
    element, if we do not take into account the replica flag and the
    visibility flag
    """

    if args:
      arguments = args[0]
      if not enoughParams(arguments):
        if isinstance(func, staticmethod):
          if hasattr(func, '__func__'):
            funcName = func.__func__.func_name
          else:
            # we may do not know the type of the method
            funcName = repr(func)
        else:
          funcName = func.func_name

        res = self.getRemoteCredentials()
        userName = res.get('username', 'UNKNOWN')
        address = self.email
        subject = '%s method!' % funcName
        body = '%s user has not provided enough input parameters! \n \
                the input parameters:%s ' % (userName, str(arguments))
        NotificationClient().sendMail(address, subject, body, 'lhcb-bookkeeping@cern.ch')
        gLogger.error('Got you: %s ---> %s' % (userName, str(arguments)))
        if self.forceExecution:  # we can force to execute the methods even the user does not
          # provide enough parameter
          result = func(self, *args)
          return result
        else:
          return S_ERROR("Provide more parameters %s" % str(arguments))  # TODO: use errno.EINVAL
      else:
        result = func(self, *args)
        return result
    else:
      result = func(self, *args)
      return result
  return checkMethodArguments



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Actual engine for adding a DM transformation, called by dirac-dms-add-
Transformation."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
__RCSID__ = "$Id$"

import os
import json
from collections import defaultdict

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.List import breakListIntoChunks

from LHCbDIRAC.TransformationSystem.Client.Transformation import Transformation
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import getProcessingPasses, BKQuery
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Utilities.PluginUtilities import getRemovalPlugins, getReplicationPlugins
from LHCbDIRAC.DataManagementSystem.Utilities.FCUtilities import chown
from LHCbDIRAC.DataManagementSystem.Client.DMScript import ProgressBar


def executeAddTransformation(pluginScript):
  """Method for actually adding a DM transformation It takes its options and
  argument values from pluginScript."""
  test = False
  start = False
  force = False
  invisible = False
  fcCheck = True
  unique = False
  bkQuery = None
  depth = None
  userGroup = None
  listProcessingPasses = False
  mcVersionSet = None
  nameOption = None

  switches = Script.getUnprocessedSwitches()
  for opt, val in switches:
    if opt in ('s', 'Start'):
      start = True
    elif opt == 'Test':
      test = True
    elif opt == "Force":
      force = True
    elif opt == "SetInvisible":
      invisible = True
    elif opt == "NoFCCheck":
      fcCheck = False
    elif opt == "Unique":
      unique = True
    elif opt == 'Chown':
      userGroup = val.split('/')
      if len(userGroup) != 2 or not userGroup[1].startswith('lhcb_'):
        gLogger.fatal("Wrong user/group")
        DIRAC.exit(2)
    elif opt == "Depth":
      try:
        depth = int(val)
      except ValueError:
        gLogger.fatal("Illegal integer depth:", val)
        DIRAC.exit(2)
    elif opt == 'ListProcessingPasses':
      listProcessingPasses = True
    elif opt == 'Name':
      nameOption = val
    elif opt == 'MCVersion':
      mcVersionSet = set(x.lower() for x in val.split(','))

  if userGroup:
    from DIRAC.Core.Security.ProxyInfo import getProxyInfo
    res = getProxyInfo()
    if not res['OK']:
      gLogger.fatal("Can't get proxy info", res['Message'])
      exit(1)
    properties = res['Value'].get('groupProperties', [])
    if 'FileCatalogManagement' not in properties:
      gLogger.error("You need to use a proxy from a group with FileCatalogManagement")
      exit(5)

  plugin = pluginScript.getOption('Plugin')
  # Default plugin for MCDST replication
  if mcVersionSet and not plugin:
    plugin = 'LHCbMCDSTBroadcastRandom'

  if not plugin and not listProcessingPasses:
    gLogger.fatal("ERROR: No plugin supplied...")
    Script.showHelp(exitCode=1)
  prods = pluginScript.getOption('Productions')
  requestID = pluginScript.getOption('RequestID')
  fileType = pluginScript.getOption('FileType')
  pluginParams = pluginScript.getPluginParameters()
  pluginSEParams = pluginScript.getPluginSEParameters()
  requestedLFNs = pluginScript.getOption('LFNs')

  transType = None
  if plugin in getRemovalPlugins():
    transType = "Removal"
  elif plugin in getReplicationPlugins():
    transType = "Replication"
  elif not listProcessingPasses:
    gLogger.notice("This script can only create Removal or Replication plugins")
    gLogger.notice("Replication :", str(getReplicationPlugins()))
    gLogger.notice("Removal     :", str(getRemovalPlugins()))
    gLogger.notice("If needed, ask for adding %s to the known list of plugins" % plugin)
    DIRAC.exit(2)

  bk = BookkeepingClient()
  tr = TransformationClient()

  if plugin in ("DestroyDataset", 'DestroyDatasetWhenProcessed') or prods:
    # visible = 'All'
    fcCheck = False

  # If mcVersions are defined, get the processing passes
  if mcVersionSet:
    # We keep only numerical values or "all"
    mcVersions = set(mcv for mcv in mcVersionSet if mcv.isdigit() or mcv == 'all')
    if not mcVersions:
      gLogger.fatal("Invalid MC versions", ','.join(sorted(mcVersionSet)))
      DIRAC.exit(2)
    if mcVersions != mcVersionSet:
      gLogger.notice("WARNING: list of MC versions reduced to", ','.join(sorted(mcVersions)))
    # Force transformations names to be unique
    unique = True
    res = ProductionRequestClient().getProductionRequestList(
        0, 'RequestID', 'DESC', 0, 0,
        {'RequestState': 'Active', 'RequestType': 'Simulation'}
    )
    if not res['OK']:
      gLogger.fatal("Error getting production requests", res['Message'])
    prodReq = res['Value']['Rows']
    reqDict = defaultdict(set)
    for row in prodReq:
      mcVersion = json.loads(row['Extra'])['mcConfigVersion']
      # If "all" we take all numerical values, otherwise we take the explicit values
      if ('all' in mcVersions and mcVersion.isdigit()) or mcVersion in mcVersions:
        simVersion = row['ProPath'].split('/')[0]
        reqDict[mcVersion].add(simVersion)
    bkPaths = []
    for mcVersion, simVersions in reqDict.items():
      for simVersion in simVersions:
        bkPaths.append('/MC/%s//%s/*Reco*' % (mcVersion, simVersion))
  elif not requestedLFNs:
    # If BK path is given, set it such that one goes through the loop below
    bkPaths = [None]
  else:
    bkPaths = []

  transBKQuery = {}
  bkQueries = []
  bkPaths.sort()
  for bkPath in sorted(bkPaths):
    if bkPath:
      bkQuery = BKQuery(bkPath)
      if not listProcessingPasses:
        gLogger.notice("For BK path: %s" % bkPath)
    else:
      bkQuery = pluginScript.getBKQuery()
    if not bkQuery and not force:
      gLogger.fatal("No LFNs and no BK query were given...")
      Script.showHelp(exitCode=2)
    if bkQuery:
      processingPass = bkQuery.getProcessingPass()
      if '...' in processingPass or '*' in processingPass:
        bkPath = bkQuery.getPath().replace('RealData', 'Real Data')
        if listProcessingPasses:
          gLogger.notice("List of processing passes for BK path", bkPath)
        processingPasses = getProcessingPasses(bkQuery, depth=depth)
        if processingPasses:
          if not listProcessingPasses:
            gLogger.notice("Transformations will be launched for the following list of processing passes:")
          gLogger.notice('\t' + '\n\t'.join(processingPasses))
        else:
          gLogger.notice("No processing passes matching the BK path")
          continue
        # Create a list of BK queries, taking into account visibility and if needed excluded file types
        exceptTypes = bkQuery.getExceptFileTypes()
        visible = bkQuery.isVisible()
        for pp in processingPasses:
          query = BKQuery(bkPath.replace(processingPass, pp), visible=visible)
          query.setExceptFileTypes(exceptTypes)
          bkQueries.append(query)
      else:
        # Single BK query
        bkQueries = [bkQuery]
    if bkPath:
      gLogger.notice("====================================")

  if listProcessingPasses:
    DIRAC.exit(0)
  reqID = pluginScript.getRequestID()
  if not requestID and reqID:
    requestID = reqID

  transGroup = plugin
  # If no BK queries are given, set to None to go once in the loop
  if not bkQueries:
    bkQueries = [None]
  for bkQuery in bkQueries:
    if bkQuery != bkQueries[0]:
      gLogger.notice("**************************************")
    # Create the transformation
    transformation = Transformation()
    transformation.setType(transType)
    transName = transType

    # In case there is a loop on processing passes
    if bkQuery:
      transBKQuery = bkQuery.getQueryDict()
    if requestedLFNs:
      longName = transGroup + " for %d LFNs" % len(requestedLFNs)
      transName += '-LFNs'
    elif prods:
      if not fileType:
        fileType = ["All"]
      prodsStr = ','.join(str(p) for p in prods)
      fileStr = ','.join(fileType)
      longName = transGroup + " of " + fileStr + " for productions %s " % prodsStr
      if len(prods) > 5:
        prodsStr = '%d-productions' % len(prods)
      if len(fileStr) > 30:
        fileStr = '%d-fileTypes' % len(fileType)
      transName += '-' + fileStr + '-' + prodsStr
    elif transBKQuery and 'FileType' in transBKQuery and \
            'BKPath' not in pluginScript.getOptions():
      if isinstance(transBKQuery['FileType'], list):
        strQuery = ','.join(transBKQuery['FileType'])
      else:
        strQuery = str(transBKQuery['FileType'])
      longName = transGroup + " for fileType " + strQuery
      transName += '-' + str(transBKQuery['FileType'])
    elif bkQuery:
      queryPath = bkQuery.getPath()
      longName = transGroup + " for BKQuery " + queryPath
      transName += '-' + queryPath
    else:
      transName = ''

    dqFlag = transBKQuery.get('DataQuality', [])
    if 'BAD' in dqFlag:
      dqFlag = ' (DQ: %s)' % ','.join(dqFlag)
      transName += dqFlag
      longName += dqFlag

    if requestID:
      transName += '-Request%s' % (requestID)
    # If a name is given in the options, use it
    if nameOption:
      transName = nameOption
      longName = transGroup + ' - ' + transName
    if not transName:
      gLogger.fatal("Didn't manage to find a name for this transformation, check options")
      DIRAC.exit(1)
    # Find a name for this transformation (transName remains the base name)
    tName = transName
    giveUp = False
    trial = 0
    while True:
      # Check if there is already a transformation with that name
      res = tr.getTransformation(tName)
      if res['OK']:
        # Transformation already exists
        if unique:
          # If unique is required and the transformation is not in a final status, give up
          if res['Value']['Status'] not in ('Archived', 'Cleaned', 'Cleaning',
                                            'Deleted', 'TransformationCleaned'):
            giveUp = True
            gLogger.notice("Transformation %s already exists with ID %d, status %s"
                           % (transName,
                              res['Value']['TransformationID'],
                              res['Value']['Status']))
            break
        trial += 1
        # Check again with new name
        tName = transName + "-" + str(trial)
      else:
        # Transformation doesn't exist, OK
        break
    # If needed, skip this BK query
    if giveUp:
      continue

    transformation.setTransformationName(tName)
    transformation.setTransformationGroup(transGroup)
    transformation.setDescription(longName[:255])
    transformation.setLongDescription(longName)
    transformation.setType(transType)
    transBody = None
    if transType == "Removal":
      if plugin == "DestroyDataset":
        transBody = "removal;RemoveFile"
      elif plugin == "DestroyDatasetWhenProcessed":
        plugin = "DeleteReplicasWhenProcessed"
        transBody = "removal;RemoveFile"
        # Set the polling period to 0 if not defined
        pluginParams.setdefault('Period', 0)
      else:
        transBody = "removal;RemoveReplica"
      transformation.setBody(transBody)

    if pluginSEParams:
      for key, val in pluginSEParams.items():
        res = transformation.setSEParam(key, val)
        if not res['OK']:
          gLogger.error('Error setting SE parameter', res['Message'])
          DIRAC.exit(1)
    if pluginParams:
      for key, val in pluginParams.items():
        res = transformation.setAdditionalParam(key, val)
        if not res['OK']:
          gLogger.error('Error setting additional parameter', res['Message'])
          DIRAC.exit(1)

    transformation.setPlugin(plugin)

    if test:
      gLogger.notice("Transformation type:", transType)
      gLogger.notice("Transformation Name:", transName)
      gLogger.notice("Transformation group:", transGroup)
      gLogger.notice("Long description:", longName)
      gLogger.notice("Transformation body:", transBody)
      if transBKQuery:
        gLogger.notice("BK Query:", transBKQuery)
      elif requestedLFNs:
        gLogger.notice("List of %d LFNs" % len(requestedLFNs))
      else:
        # Should not happen here, but who knows ;-)
        gLogger.error("No BK query provided...")
        Script.showHelp(exitCode=1)

    if force:
      lfns = []
    elif transBKQuery:
      bkPath = bkQuery.getPath()
      title = "Executing BK query"
      if bkPath:
        title += " for " + bkPath
      else:
        title += ": " + str(bkQuery)
      progressBar = ProgressBar(1, title=title)
      lfns = bkQuery.getLFNs(printSEUsage=(transType == 'Removal' and
                                           not pluginScript.getOption('Runs')),
                             printOutput=test)
      progressBar.endLoop(message=("found %d files" % len(lfns) if lfns else 'no files found'))
    else:
      lfns = requestedLFNs
    nfiles = len(lfns)

    if test:
      gLogger.notice("Plugin:", plugin)
      gLogger.notice("Parameters:", pluginParams)
      gLogger.notice("RequestID:", requestID)
      continue

    if not force and nfiles == 0:
      gLogger.notice("No files, but if you anyway want to submit the transformation, use option --Force")
      continue

    if userGroup:
      directories = set(os.path.dirname(lfn) for lfn in lfns)
      res = chown(directories, user=userGroup[0], group=userGroup[1])
      if not res['OK']:
        gLogger.fatal("Error changing ownership", res['Message'])
        DIRAC.exit(3)
      gLogger.notice("Successfully changed owner/group for %d directories" % res['Value'])

    # If the transformation is a removal transformation,
    #  check all files are in the FC. If not, remove their replica flag
    if fcCheck and transType == 'Removal':
      from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
      fc = FileCatalog()
      success = 0
      missingLFNs = set()
      chunkSize = 1000
      progressBar = ProgressBar(len(lfns), title="Checking %d files in FC" % len(lfns), chunk=chunkSize)
      for lfnChunk in breakListIntoChunks(lfns, chunkSize):
        progressBar.loop()
        res = fc.exists(lfnChunk)
        if res['OK']:
          success += len(res['Value']['Successful'])
          missingLFNs |= set(res['Value']['Failed'])
        else:
          gLogger.fatal("\nError checking files in the FC", res['Message'])
          DIRAC.exit(2)
      progressBar.endLoop(message='all found' if not missingLFNs else None)
      if missingLFNs:
        gLogger.notice('%d are in the FC, %d are not. Attempting to remove GotReplica' % (success, len(missingLFNs)))
        res = bk.removeFiles(list(missingLFNs))
        if res['OK']:
          gLogger.notice("Replica flag successfully removed in BK")
        else:
          gLogger.fatal("Error removing BK flag", res['Message'])
          DIRAC.exit(2)

    # Prepare the transformation
    if transBKQuery:
      # !!!! Hack in order to remain compatible with hte BKQuery table...
      for transKey, bkKey in (('ProductionID', 'Production'),
                              ('RunNumbers', 'RunNumber'),
                              ('DataQualityFlag', 'DataQuality')):
        if bkKey in transBKQuery:
          transBKQuery[transKey] = transBKQuery.pop(bkKey)
      transformation.setBkQuery(transBKQuery)

    # If the transformation uses the RemoveDatasetFromDisk plugin, set the files invisible in the BK...
    # Try and let them visible such that users can see they are archived...
    # It was:
    # setInvisiblePlugins = ("RemoveDatasetFromDisk", )
    setInvisiblePlugins = tuple()
    if invisible or plugin in setInvisiblePlugins:
      chunkSize = 1000
      progressBar = ProgressBar(len(lfns), title='Setting %d files invisible' % len(lfns), chunk=chunkSize)
      okFiles = 0
      for lfnChunk in breakListIntoChunks(lfns, chunkSize):
        progressBar.loop()
        res = bk.setFilesInvisible(lfnChunk)
        if res['OK']:
          okFiles += len(lfnChunk)
      if okFiles == len(lfns):
        msg = "all files successfully set invisible in BK"
      else:
        msg = "%d files successfully set invisible in BK" % okFiles
      progressBar.endLoop(message=msg)
      if res['OK']:
        if transBKQuery:
          savedVisi = transBKQuery.get('Visible')
          transBKQuery['Visible'] = 'All'
          transformation.setBkQuery(transBKQuery.copy())
          if savedVisi:
            transBKQuery['Visible'] = savedVisi
          else:
            transBKQuery.pop('Visible')
      else:
        gLogger.error("Failed to set files invisible: ", res['Message'])

    errMsg = ''
    while True:
      res = transformation.addTransformation()
      if not res['OK']:
        errMsg = "Couldn't create transformation"
        break
      res = transformation.getTransformationID()
      if res['OK']:
        transID = res['Value']
      else:
        errMsg = "Error getting transformationID"
        break
      # If some LFNs must be added, do it now
      if requestedLFNs:
        from LHCbDIRAC.TransformationSystem.Utilities.PluginUtilities import addFilesToTransformation
        res = addFilesToTransformation(transID, requestedLFNs, addRunInfo=True)
        if not res['OK']:
          errMsg = "Could not add files to transformation"
          break
        gLogger.notice("%d files successfully added to transformation" % len(res['Value']))
      if requestID:
        transformation.setTransformationFamily(requestID)
      if start:
        transformation.setStatus('Active')
        transformation.setAgentType('Automatic')
      gLogger.notice("Name: %s" % transName)
      gLogger.notice("Transformation body:", transBody)
      gLogger.notice("Plugin:", plugin)
      if pluginParams:
        gLogger.notice("Additional parameters:", pluginParams)
      if requestID:
        gLogger.notice("RequestID:", requestID)
      break
    if errMsg:
      gLogger.notice(errMsg, res['Message'])

  DIRAC.exit(0)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Insert new file types in the Bookkeeping."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from builtins import input

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile]' % Script.scriptName]))
  Script.parseCommandLine(ignoreErrors=True)

  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  bk = BookkeepingClient()

  exitCode = 0

  ftype = input("FileType: ")
  desc = input("Description: ")
  version = input("File type version: ")
  gLogger.notice('Do you want to add this new file type? (yes or no)')
  value = input('Choice:')
  choice = value.lower()
  if choice in ['yes', 'y']:
    res = bk.insertFileTypes(ftype.upper(), desc, version)
    if res['OK']:
      gLogger.notice('The file types added successfully!')
    else:
      gLogger.error("Error discovered!", res['Message'])
  elif choice in ['no', 'n']:
    gLogger.notice('Aborted!')
  else:
    gLogger.error('Unexpected choice:', value)

  DIRAC.exit(exitCode)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: RunDBInterfaceHandler

.. module: RunDBInterfaceHandler

:synopsis: DISET request handler base class for the DatasetDB
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time
import sys

import six

# from DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler

__RCSID__ = "$Id$"

allRunFields = ['runID', 'fillID', 'state', 'runType', 'partitionName', 'partitionID',
                'startTime', 'endTime', 'destination', 'startLumi', 'endLumi', 'beamEnergy']
allFileFields = ['fileID', 'runID', 'name', 'state', 'bytes',
                 'events', 'stream', 'creationTime', 'timeStamp', 'refCount']

server = None
runStates = {}
runStateRev = {}
fileStates = {}
fileStateRev = {}


def initializeRunDBInterfaceHandler(serviceInfo):
  global server
  # sys.path.insert(0, '/home/rainer/projects/RunDatabase/python')
  # sys.path.append( '/group/online/rundb/RunDatabase/python' )
  sys.path.append('/admin/RunDatabase/python')
  from path import SQL_ALCHEMY_PATH  # pylint: disable=import-error,no-name-in-module
  sys.path.append(SQL_ALCHEMY_PATH)
  try:
    ORACLE_HOME = os.environ['ORACLE_HOME']
  except BaseException:
    print('ERROR: ORACLE_HOME environment variable should be set')
  sys.path.append("%s" % str(ORACLE_HOME))
  import RunDatabase_Defines  # pylint: disable=import-error
  # print dir(RunDatabase_Defines)
  # print RunDatabase_Defines.FileFields

  from RunDatabase_Defines import RUN_STATE_TRANSLATION, FILE_STATE_TRANSLATION  # pylint: disable=import-error

  global fileStates
  global fileStateRev
  fileStates = FILE_STATE_TRANSLATION
  fileStateRev = {}
  for key, value in fileStates.items():
    fileStateRev[value] = key
  global runStates
  global runStateRev
  runStates = RUN_STATE_TRANSLATION
  runStateRev = {}
  for key, value in runStates.items():
    runStateRev[value] = key

  im = __import__('RunDatabase', globals(), locals(), ['*'])
  import RunDatabase  # pylint: disable=import-error
  print(im)
  from DbModel import createEngine_Oracle  # pylint: disable=import-error
  try:
    server = RunDatabase.RunDbServer(engine=createEngine_Oracle())
  except BaseException:
    print('Failed to make an instance of runDB server')
  return S_OK()


class RunDBInterfaceHandler(RequestHandler):
  """
  .. class:: RunDBInterfaceHandler
  """
  types_getFilesSummaryWeb = [dict, list, int, int]

  def export_getFilesSummaryWeb(self, selectDict, sortList, startItem, maxItems):
    """export of getFilesSummaryWeb."""
    paramString = ''
    for selectParam in allFileFields:
      if selectParam in selectDict:
        selectValue = selectDict[selectParam]
        if selectParam == 'state':
          intStates = []
          for strState in selectValue:
            intStates.append(fileStateRev[strState])
          selectValue = intStates
        if isinstance(selectValue, six.string_types):
          paramString = "%s,%s='%s'" % (paramString, selectParam, selectValue)
        else:
          paramString = "%s,%s=%s" % (paramString, selectParam, selectValue)
    decending = False
    if sortList:
      paramString = "%s,orderBy='%s'" % (paramString, sortList[0][0])
      if sortList[0][1] == 'DESC':
        decending = True
    paramString = "%s,no=%s" % (paramString, sys.maxsize)
    if paramString:
      filesQueryString = "success,result = server.getFilesDirac(fields=allFileFields%s)" % paramString
    else:
      filesQueryString = "success,result = server.getFilesDirac(fields=allFileFields)"
    print(filesQueryString)
    exec(filesQueryString)
    if not success:  # using exec statement above -> pylint: disable=undefined-variable
      return S_ERROR(result)  # using exec statement above -> pylint: disable=undefined-variable
    resultDict = {}
    nFiles = len(result)  # using exec statement above -> pylint: disable=undefined-variable
    resultDict['TotalRecords'] = nFiles
    if nFiles == 0:
      return S_OK(resultDict)
    if decending:
      result.reverse()  # using exec statement above -> pylint: disable=undefined-variable

    statusCountDict = {}
    for res in result:  # using exec statement above -> pylint: disable=undefined-variable
      state = res[3]
      if state in fileStates:
        state = fileStates[state]
      else:
        state = 'UNKNOWN'
      if state not in statusCountDict:
        statusCountDict[state] = 0
      statusCountDict[state] += 1
    resultDict['Extras'] = statusCountDict

    iniFile = startItem
    lastFile = iniFile + maxItems
    if iniFile >= nFiles:
      return S_ERROR('Item number out of range')
    if lastFile > nFiles:
      lastFile = nFiles
    fileList = result[iniFile:lastFile]  # using exec statement above -> pylint: disable=undefined-variable

    # prepare the standard structure now
    resultDict['ParameterNames'] = allFileFields
    records = []
    for fileTuple in fileList:
      ['fileID', 'runID', 'name', 'state', 'bytes', 'events', 'stream', 'creationTime', 'timeStamp', 'refCount']
      fileID, runID, name, state, bytes, events, stream, creationTime, timeStamp, refCount = fileTuple
      timeStamp = str(timeStamp)
      creationTime = str(creationTime)
      if state in fileStates:
        state = fileStates[state]
      else:
        state = 'UNKNOWN'
      records.append((fileID, runID, name, state, bytes, events, stream, creationTime, timeStamp, refCount))

    resultDict['Records'] = records
    print(resultDict)
    return S_OK(resultDict)

  """
  getFiles(self, fields        = ['name'],
                     fileID        = None,
                     runID         = None,
                     name          = None,
                     stream        = None,
                     state         = None,
                     timeout       = None,
                     refCount      = None,
                     runType       = None,
                     runPartName   = None,
                     runPartID     = None,
                     runStartTime  = None,
                     runEndTime    = None,
                     runDest       = None,
                     runState      = None,
                     orderBy       = None,
                     no            = 100
              ):
  """

  types_getFileSelections = []

  def export_getFileSelections(self):
    pass

  types_getRunsSummaryWeb = [dict, list, int, int]

  def export_getRunsSummaryWeb(self, selectDict, sortList, startItem, maxItems):
    """export of getRunsSummaryWeb."""
    paramString = ''
    for selectParam in allRunFields:
      if selectParam in selectDict:
        selectValue = selectDict[selectParam]
        if selectParam == 'state':
          intStates = []
          for strState in selectValue:
            intStates.append(runStateRev[strState])
          selectValue = intStates
        if isinstance(selectValue, six.string_types):
          paramString = "%s,%s='%s'" % (paramString, selectParam, selectValue)
        else:
          paramString = "%s,%s=%s" % (paramString, selectParam, selectValue)
    decending = False
    if sortList:
      paramString = "%s,orderBy='%s'" % (paramString, sortList[0][0])
      if sortList[0][1] == 'DESC':
        decending = True
    paramString = "%s,no=%s" % (paramString, sys.maxsize)
    jobsQueryString = "success,result = "
    jobsQueryString += "server.getRunsDirac("
    jobsQueryString += "fields=allRunFields,runExtraParams=['magnetCurrent','magnetState']"
    if paramString:
      jobsQueryString += "%s" % paramString
    jobsQueryString += ")"
    print(jobsQueryString)
    exec(jobsQueryString)
    if not success:  # using exec statement above -> pylint: disable=E0601
      return S_ERROR(result)  # using exec statement above -> pylint: disable=E0601
    resultDict = {}
    nRuns = len(result)
    resultDict['TotalRecords'] = nRuns
    if nRuns == 0:
      return S_OK(resultDict)

    if decending:
      result.reverse()

    statusCountDict = {}
    for res in result:
      state = res[2]
      if state in runStates:
        state = runStates[state]
      else:
        state = 'UNKNOWN'
      if state not in statusCountDict:
        statusCountDict[state] = 0
      statusCountDict[state] += 1
    resultDict['Extras'] = statusCountDict

    iniRun = startItem
    lastRun = iniRun + maxItems
    if iniRun >= nRuns:
      return S_ERROR('Item number out of range')
    if lastRun > nRuns:
      lastRun = nRuns
    runList = result[iniRun:lastRun]

    # prepare the standard structure now
    runCounters = {}
    for runTuple in runList:
      (
          runID,
          fillID,
          state,
          runType,
          partitionName,
          partitionID,
          startTime,
          endTime,
          destination,
          startLumi,
          endLumi,
          beamEnergy,
          magnetCurrent,
          magnetState,
      ) = runTuple
      runCounters[runID] = {'Size': 0, 'Events': 0, 'Files': 0}

    # Now sum the number of events and files in the run
    if runCounters:
      success, result = server.getFilesDirac(
          fields=allFileFields, runID=list(runCounters), orderBy='runID', no=sys.maxsize)
      if not success:
        return S_ERROR(result)
      for item in result:
        runID = item[1]
        size = item[4]
        events = item[5]
        if size and events:
          runCounters[runID]['Size'] += size
          runCounters[runID]['Events'] += events
          runCounters[runID]['Files'] += 1

    totalFiles = 0
    totalEvents = 0
    totalSize = 0
    for runID in runCounters:
      totalFiles += runCounters[runID]['Files']
      totalEvents += runCounters[runID]['Events']
      totalSize += runCounters[runID]['Size']
      size = "%.2f" % (runCounters[runID]['Size'] / (1000 * 1000 * 1000.0))
      runCounters[runID]['Size'] = size
    resultDict['Counters'] = {'Files': totalFiles, 'Events': totalEvents,
                              'Size': "%.2f" % (totalSize / (1000 * 1000 * 1000.0))}

    records = []
    for runTuple in runList:
      (
          runID,
          fillID,
          state,
          runType,
          partitionName,
          partitionID,
          startTime,
          endTime,
          destination,
          startLumi,
          endLumi,
          beamEnergy,
          magnetCurrent,
          magnetState
      ) = runTuple
      startTime = str(startTime)
      endTime = str(endTime)
      if state in runStates:
        state = runStates[state]
      else:
        state = 'UNKNOWN'
      try:
        integratedLumi = endLumi - startLumi
      except BaseException:
        integratedLumi = 'na'
      records.append(
          (runID,
           fillID,
           state,
           runType,
           partitionName,
           partitionID,
           startTime,
           endTime,
           destination,
           startLumi,
           endLumi,
           beamEnergy,
           runCounters[runID]['Files'],
           runCounters[runID]['Events'],
           runCounters[runID]['Size'],
           magnetCurrent,
           magnetState,
           integratedLumi))
    resultDict['Records'] = records
    resultDict['ParameterNames'] = allRunFields + ['files', 'events',
                                                   'size', 'magnetCurrent', 'magnetState', 'integratedLumi']
    print('parameter names: ', resultDict['ParameterNames'])
    print(resultDict)
    return S_OK(resultDict)

  types_getRunSelections = []

  def export_getRunSelections(self):
    """export of getRunSelections."""
    try:
      paramDict = {}

      queries = [('PartitionName', 'getPartitionNames'),
                 ('RunType', 'getRunTypes'),
                 ('FillID', 'getFillIDs'),
                 ('Destination', 'getDestinations'),
                 ('StartLumi', 'getStartLumis'),
                 ('EndLumi', 'getEndLumis'),
                 ('BeamEnergy', 'getBeamEnergies')]

      for key, query in queries:
        startTime = time.time()
        execString = "success,result = server.%s()" % query
        print(execString)
        exec(execString)  # pylint: disable=exec-used
        gLogger.debug(
            "RunDBInterfaceHandler.getSelections: server.%s() took %.2f seconds." %
            (query, time.time() - startTime))
        if not success:  # using exec statement above -> pylint: disable=used-before-assignment
          errStr = "RunDBInterfaceHandler.getSelections: Failed to get distinct %s." % key
          gLogger.error(errStr, result)  # using exec statement above -> pylint: disable=E0601
          return S_ERROR(errStr)
        paramDict[key] = [res for res in result if res]

      startTime = time.time()
      success, result = server.getRunStates()
      gLogger.debug(
          "RunDBInterfaceHandler.getSelections: server.getRunStates() took %.2f seconds." %
          (time.time() - startTime))
      if not success:
        errStr = "RunDBInterfaceHandler.getSelections: Failed to get distinct run States."
        gLogger.error(errStr, result)
        return S_ERROR(errStr)
      states = []
      for runStat in result:
        if runStat in runStates:
          states.append(runStates[runStat])
      paramDict['State'] = states
      return S_OK(paramDict)
    except Exception as x:
      errStr = "RunDBInterfaceHandler.getSelections: Exception while obtaining possible run configurations."
      gLogger.exception(errStr, '', x)
      return S_ERROR("%s %s" % (errStr, x))

  types_getRunParams = [int]

  def export_getRunParams(self, runID):
    """export of getRunParams."""
    success, result = server.getRunParams(runID)
    if not success:
      return S_ERROR(result)
    return S_OK(result)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""File utilities module (e.g. make GUIDs)"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import uproot

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.File import makeGuid as DIRACMakeGUID


def getRootFileGUIDs(fileList):
  """Retrieve a list of GUIDs for a list of files."""
  guids = {'Successful': {}, 'Failed': {}}
  for fileName in fileList:
    res = getRootFileGUID(fileName)
    if res['OK']:
      gLogger.verbose('GUID from ROOT', '%s' % res['Value'])
      guids['Successful'][fileName] = res['Value']
    else:
      guids['Failed'][fileName] = res['Message']
  return S_OK(guids)


def getRootFileGUID(fileName):
  """Function to retrieve a file GUID using uproot."""
  try:
    f = uproot.open(fileName)
    branch = f['Refs']['Params']
    for item in branch.array():
      item = item.decode()
      if item.startswith('FID='):
        return S_OK(item.split('=')[1])
    return S_ERROR('GUID not found')
  except Exception as e:
    errorMsg = 'Error extracting GUID: %s' % e
    return S_ERROR(errorMsg)


def makeGuid(fileNames):
  """Function to retrieve a file GUID using uproot."""
  if isinstance(fileNames, six.string_types):
    fileNames = [fileNames]

  fileGUIDs = {}
  for fileName in fileNames:
    res = getRootFileGUID(fileName)
    if res['OK']:
      gLogger.verbose('GUID from ROOT', '%s' % res['Value'])
      fileGUIDs[fileName] = res['Value']
    else:
      gLogger.info('Could not obtain GUID from file through Gaudi, using standard DIRAC method')
      fileGUIDs[fileName] = DIRACMakeGUID(fileName)

  return fileGUIDs



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.ResourceStatusSystem.Policy.TransferQualityPolicy.

TransferQualityPolicy.__bases__:
  DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase.PolicyBase
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase import PolicyBase

__RCSID__ = "$Id$"


class TransferQualityPolicy(PolicyBase):
  """The TransferQualityPolicy class is a policy class to check the transfer
  quality.

  Evaluates the TransferQuality results given by the
  DIRACAccounting.TransferQuality command against a certain set of
  thresholds defined in the CS.
  """

  @staticmethod
  def _evaluate(commandResult):
    """Evaluate policy on Data quality.

    :returns:
        {
          'Status':Error|Unknown|Active|Probing|Banned,
          'Reason':'TransferQuality:None'|'TransferQuality:xx%',
        }
    """

    result = {
        'Status': None,
        'Reason': None
    }

    if not commandResult['OK']:
      result['Status'] = 'Error'
      result['Reason'] = commandResult['Message']
      return S_OK(result)

    commandResult = commandResult['Value']

    if not commandResult:
      result['Status'] = 'Unknown'
      result['Reason'] = 'No values to take a decision'
      return S_OK(result)

    if 'Name' not in commandResult:
      result['Status'] = 'Error'
      result['Reason'] = 'Missing "Name" key'
      return S_OK(result)

    name = commandResult['Name']

    if 'Mean' not in commandResult:
      result['Status'] = 'Error'
      result['Reason'] = 'Missing "Mean" key'
      return S_OK(result)

    mean = commandResult['Mean']

    if mean is None:
      result['Status'] = 'Unknown'
      result['Reason'] = 'No values to take a decision'
      return S_OK(result)

    result['Reason'] = 'TransferQuality: %d -> ' % mean

    # FIXME: policyParameters = Configurations.getPolicyParameters()

    policyParameters = {
        'Transfer_QUALITY_LOW': 60,
        'Transfer_QUALITY_HIGH': 90
    }

    if 'FAILOVER'.lower() in name.lower():

      if mean < policyParameters['Transfer_QUALITY_LOW']:
        result['Status'] = 'Degraded'
        strReason = 'Low'
      elif mean < policyParameters['Transfer_QUALITY_HIGH']:
        result['Status'] = 'Active'
        strReason = 'Mean'
      else:
        result['Status'] = 'Active'
        strReason = 'High'

    else:

      if mean < policyParameters['Transfer_QUALITY_LOW']:
        result['Status'] = 'Degraded'
        strReason = 'Low'
      elif mean < policyParameters['Transfer_QUALITY_HIGH']:
        result['Status'] = 'Degraded'
        strReason = 'Mean'
      else:
        result['Status'] = 'Active'
        strReason = 'High'

    result['Reason'] = result['Reason'] + strReason
    return S_OK(result)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""List file types from the Bookkeeping."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile]' % Script.scriptName]))
  Script.parseCommandLine(ignoreErrors=True)

  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  bk = BookkeepingClient()
  exitCode = 0

  res = bk.getAvailableFileTypes()

  if res['OK']:
    dbresult = res['Value']
    print('Filetypes:')
    for record in dbresult['Records']:
      print(str(record[0]).ljust(30) + str(record[1]))

  DIRAC.exit(exitCode)


if __name__ == "__main__":
  main()
